import argparse
import os
from typing import Generator, Any, Set


def unique(seq: [Any]) -> [Any]:
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def read_all_lines(filename: str) -> [str]:
    with open(filename) as file:
        lines = file.readlines()
    return [line.strip() for line in lines]


def files_by_extension(directory: str, extension: str) -> Generator[str, None, None]:
    for root, dirs, files in os.walk(directory):
        for file in files:
            if file.endswith(extension):
                yield os.path.join(root, file)


def get_dependencies(directory: str) -> [str]:
    def get_dependencies_rec(directory_rec: str):
        deps = None
        deps_rec = []
        result = None
        dep_file = os.path.join(directory_rec, "dependencies.txt")
        if os.path.isfile(dep_file):
            with open(dep_file) as f:
                deps = [line.strip() for line in f.readlines() if len(line.strip()) > 0]
        else:
            deps = []

        for dep in deps:
            deps_rec.extend(get_dependencies_rec(os.path.join(directory_rec, "..", dep)))

        result = deps_rec
        result.extend(deps)
        return result

    return unique(get_dependencies_rec(directory))


def get_files(source_dir: str, target_dir: str, outdated: bool) -> [str]:
    files = []
    excluded_packages = []
    exclude_txt_file = os.path.join(source_dir, "..", "exclude.txt")
    if os.path.isfile(exclude_txt_file):
        excluded_packages = read_all_lines(exclude_txt_file)
    excluded_dirs = get_excluded_directories(source_dir, excluded_packages)

    for source_file in files_by_extension(source_dir, ".java"):
        skip=False
        for excluded_dir in excluded_dirs:
            if source_file.startswith(excluded_dir):
                skip=True
                break
        if skip:
            continue

        if not outdated:
            files.append(source_file)
            continue
        relative_path = os.path.relpath(source_file, start=source_dir)
        target_file = os.path.splitext(os.path.join(target_dir, relative_path))[0] + ".class"
        if not os.path.isfile(target_file):
            files.append(source_file)
        elif os.path.getmtime(source_file) > os.path.getmtime(target_file):
            files.append(source_file)
    return files


def get_excluded_directories(source_dir: str, excludes: [str]) -> Set[str]:
    return set(map(lambda p: os.path.join(source_dir, p.replace(".", os.path.sep)), excludes))


def dependencies():
    for dep in get_dependencies("."):
        print(dep)


def all_files():
    files = []
    files.extend(get_files("src", None, outdated=False))
    for dep in get_dependencies("."):
        files.extend(get_files(os.path.join("..", dep, "src"), None, outdated=False))

    for file in files:
        print(file)


def outdated():
    target_dir = "build/classes"
    outdated_files = []
    outdated_files.extend(get_files("src", target_dir, outdated=True))
    for dep in get_dependencies("."):
        outdated_files.extend(get_files(os.path.join("..", dep, "src"), target_dir, outdated=True))

    for file in outdated_files:
        print(file)


def main():
    parser = argparse.ArgumentParser("utility to support java build and dependency management")
    subparsers = parser.add_subparsers()

    sp_dependencies = subparsers.add_parser("dependencies", help="print all direct and indirect dependencies")
    sp_dependencies.set_defaults(func=dependencies)

    sp_all = subparsers.add_parser("all", help="print all java files")
    sp_all.set_defaults(func=all_files)

    sp_outdated = subparsers.add_parser("outdated", help="print all outdated java files")
    sp_outdated.set_defaults(func=outdated)

    args = parser.parse_args()

    if 'func' in args:
        args.func()
    else:
        parser.print_help()


if __name__ == '__main__':
    main()
