/*
 * COSMA -- A Virtual Machine for the Secure Cooperation of Untrusted Components.
 * Copyright (C) 2019  Roland Wismüller and Damian Ludwig, University of Siegen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package cola;

import cola.ast.ClassNode;
import cola.ast.ComponentNode;
import cola.ast.TypedefNode;
import cola.compiler.ComponentGenerator;
import cola.parser.Cola;
import cosma.argparse.Arg;
import cosma.argparse.ArgParse;
import cosma.argparse.PositionalArgs;
import cosma.bytecode.builder.Builder;
import cosma.tools.logger.LogManager;
import cosma.tools.logger.Logger;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * The compilers entry point.
 */
public class Colac {
    private static final Logger logger = LogManager.getCurrentClassLogger();
    
    public static class ColacOptions {
        @Arg(name = "showDebug", shortName = 'd',
             helpText = "Print debug information during compilation.")
        public boolean showDebug;

        @Arg(name = "showWarnings", shortName = 'w',
             helpText = "Print warnings that come up during compilation.")
        public boolean showWarnings;

        @Arg(name = "output", shortName = 'o', defaultValue = "a.bcoil",
             helpText = "Output into the specified file.", paramName = "filename")
        public String output;

        @PositionalArgs(paramName = "filenames", minRequired = 1)
        public String[] filenames;
    }
    
    public static void compile(ComponentNode c, String outfile) {
        /* Generate code */
        Builder top = new Builder(true);
        ComponentGenerator componentGenerator = new ComponentGenerator(top);
        componentGenerator.generateComponent(c);

        /* Export */
        try {
            top.writeTo(outfile);
            logger.info("Component exported to {file}", outfile);
        } catch (java.io.IOException e) {
            logger.error("Error exporting component to {file}", outfile);
            System.exit(1);
        }
    }

    private static ComponentNode buildFromFiles(String[] filenames) {
        Map<String, ComponentNode> discoveredUnits = new HashMap<>();
        List<String> requiredUnits = new ArrayList<>();
        String componentName = null;

        for (int i=0;i<filenames.length;i++) {
            String filename = filenames[i];
            logger.debug("Processing file: {file}", filename);
            ComponentNode unit = Cola.parseFromFile(filename);
            logger.debug("File {filename} contains unit {unit}", filename, unit.getName());
            discoveredUnits.put(unit.getName(), unit);
            if ((componentName == null) && unit.isComponent()) {
                /* By convention, the first file must contain the
                 * principal class and therefor is always required */
                requiredUnits.add(unit.getName());
                componentName = unit.getName();
            }
        }

        /////
        if (componentName == null) {
            logger.error("No main class found.");
            System.exit(1);
        }
        /////

        List<ClassNode> classNodes = new ArrayList<>();
        List<TypedefNode> typedefs = new ArrayList<>();
        int lastRequiredSize = 0;
        do {
            int currentRequiredSize = requiredUnits.size();
            for (;lastRequiredSize<currentRequiredSize;lastRequiredSize++) {
                String unitName = requiredUnits.get(lastRequiredSize);
                ComponentNode unit = discoveredUnits.get(unitName);
                /////
                if (unit == null) {
                    logger.error("Cannot find required unit {unit}", unitName);
                    System.exit(1);
                }
                /////
                if (unit.getClasses() != null)
                    classNodes.addAll(unit.getClasses());
                if (unit.getTypedefs() != null)
                    typedefs.addAll(unit.getTypedefs());
                if (unit.getDependencies() != null) {
                    for (String dependency : unit.getDependencies()) {
                        if (!requiredUnits.contains(dependency))
                            requiredUnits.add(dependency);
                    }
                }
            }
        } while (lastRequiredSize < requiredUnits.size());
        
        return new ComponentNode(componentName,
                                 classNodes, typedefs, null,
                                 ComponentNode.UnitType.Component);
    }

    public static void main(String[] args) {
        ColacOptions options = ArgParse.parse(ColacOptions.class, Colac.class, args);
        logger.info("Compiling file(s): {files}", java.util.Arrays.toString(options.filenames));
        ComponentNode component = buildFromFiles(options.filenames);
        compile(component, options.output);
    }
}
