/*
 * COSMA -- A Virtual Machine for the Secure Cooperation of Untrusted Components.
 * Copyright (C) 2019  Roland Wismüller and Damian Ludwig, University of Siegen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package cola.ast;

import cola.ast.stm.BlockStm;
import cola.ast.type.AbstractType;
import cola.ast.type.FunctionalType;
import cola.ast.visitor.GenericVisitor;
import cola.ast.visitor.VoidVisitor;

import java.util.List;

/**
 * This class represents a method.
 */
public class MethodNode extends Node {
    private final String name;
    private final boolean isStatic;
    private final boolean isOptional;
    private final ModifierNode modifier;
    private final List<AbstractType> resultTypes;
    private final List<IdentifierWithType> parameters;
    private final BlockStm body;

    /**
     * Constructs a new method.
     *
     * @param name the methods name
     * @param parameters the formal parameters (name and type)
     * @param resultTypes the result types
     * @param modifier the access modifier
     * @param isStatic whether the method is _static_
     * @param isOptional whether the method is optional
     * @param body the methods body statement (nullable)
     */
    public MethodNode(String name,
                      List<IdentifierWithType> parameters,
                      List<AbstractType> resultTypes,
                      ModifierNode modifier,
                      boolean isStatic,
                      boolean isOptional,
                      BlockStm body) {
        this.name = name;
        this.parameters = parameters;
        this.resultTypes = resultTypes;
        this.modifier = modifier;
        this.isStatic = isStatic;
        this.isOptional = isOptional;
        this.body = body;
    }

    /**
     * Get the methods parameter types.
     *
     * @return a list of types
     */
    public List<AbstractType> getParameterTypes() {
        if (parameters == null || parameters.size() == 0) {
            return null;
        } else {
            return parameters
                .stream()
                .map(IdentifierWithType::getType)
                .collect(java.util.stream.Collectors.toList());
        }
    }

    public String getName() {
        return name;
    }

    public List<IdentifierWithType> getParameters() {
        return parameters;
    }

    public boolean hasParameters() {
        return parameters != null;
    }

    public boolean hasResults() {
        return resultTypes != null;
    }

    public List<AbstractType> getResultTypes() {
        return resultTypes;
    }

    public ModifierNode getModifier() {
        return modifier;
    }

    public boolean isStatic() {
        return isStatic;
    }

    public boolean isOptional() {
        return isOptional;
    }

    public BlockStm getBody() {
        return body;
    }

    /**
     * Get the methods type.
     *
     * @return a (function) type
     */
    public AbstractType getType() {
        return new FunctionalType(getParameterTypes(), resultTypes);
    }

    public <T> void accept(VoidVisitor<T> visitor, T arg) {
        visitor.visit(this, arg);
    }

    public <T,R> R accept(GenericVisitor<T,R> visitor, T arg) {
        return visitor.visit(this, arg);
    }
}
