/*
 * COSMA -- A Virtual Machine for the Secure Cooperation of Untrusted Components.
 * Copyright (C) 2019  Roland Wismüller and Damian Ludwig, University of Siegen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package cola.ast.type;

import cola.ast.visitor.GenericVisitor;
import cola.ast.visitor.VoidVisitor;

/**
 * This class represents an array type.
 */
public class ArrayType extends AbstractType {
    private final AbstractType elementType;
    private final int dimensions;

    /**
     * Constructs a new array type with base type /basetype/ and /dims/
     * dimensions.
     *
     * @param elementType The type of the array's elements
     * @param dimensions The array's dimensions
     */
    public ArrayType(AbstractType elementType, int dimensions) {
        this.elementType = elementType;
        this.dimensions = dimensions;
    }

    public <T> void accept(VoidVisitor<T> visitor, T arg) {
        visitor.visit(this, arg);
    }

    public <T,R> R accept(GenericVisitor<T,R> visitor, T arg) {
        return visitor.visit(this, arg);
    }

    @Override
    public boolean equals(Object other) {
        if (other instanceof ArrayType) {
            ArrayType type = (ArrayType) other;
            return elementType.equals(type.elementType)
                && (dimensions == type.dimensions)
                && (assertions == type.assertions);
        }
        return false;
    }

    public int getDimensions() {
        return dimensions;
    }

    public AbstractType getElementType() {
        return elementType;
    }

    public String getPrettyName() {
        return elementType.getPrettyName() + "[" + dimensions + "]";
    }

    public String getInternalName() {
        return "(" + elementType.getInternalName() + ")[" + dimensions + "]" + assertionsToString();
    }

    @Override
    public int hashCode() {
        int elementTypeHash = elementType.hashCode();
        return elementTypeHash * 91 + dimensions * 17 + assertions;
    }

    public ArrayType asLocal() {
        if (!isLocal()) {
            ArrayType local = new ArrayType(elementType, dimensions);
            local.setAssertions(assertions+1);
            return local;
        }
        return this;
    }
}
