/*
 * COSMA -- A Virtual Machine for the Secure Cooperation of Untrusted Components.
 * Copyright (C) 2019  Roland Wismüller and Damian Ludwig, University of Siegen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package cola.ast.type;

import cola.ast.visitor.GenericVisitor;
import cola.ast.visitor.VoidVisitor;

import java.util.List;

/**
 * This class represents a function type.
 */
public class FunctionalType extends AbstractType {
    private static int __nextID__ = 0;
    
    private final List<AbstractType> from;
    private final List<AbstractType> to;
    private final int id;

    /**
     * Constructs a new function type.
     *
     * @param from a list of parameter types
     * @param to a list of result types
     */
    public FunctionalType(List<AbstractType> from, List<AbstractType> to) {
        this.from = from;
        this.to = to;
        this.id = __nextID__++;
    }

    public List<AbstractType> getParameterTypes() {
        return from;
    }

    public List<AbstractType> getResultTypes() {
        return to;
    }

    public boolean hasResults() {
        return to != null;
    }

    public boolean hasParameters() {
        return from != null;
    }

    public <T> void accept(VoidVisitor<T> visitor, T arg) {
        visitor.visit(this, arg);
    }

    public <T,R> R accept(GenericVisitor<T,R> visitor, T arg) {
        return visitor.visit(this, arg);
    }

    @Override
    public boolean equals(Object other) {
        if (other instanceof FunctionalType) {
            FunctionalType t = (FunctionalType) other;
            if ((hasParameters() && t.hasParameters()) && (t.from.size() == from.size())) {
                for (int i=0;i<from.size();i++) {
                    if (!from.get(i).equals(t.from.get(i)))
                        return false;
                }
            }
            if ((hasResults() && t.hasResults()) && (t.to.size() == to.size())) {
                for (int i=0;i<to.size();i++) {
                    if (!to.get(i).equals(t.to.get(i)))
                        return false;
                }
            }
            return assertions == t.assertions;
        }
        return false;
    }

    @Override
    public int hashCode() {
        final int firstPrime = 31;
        final int secondPrime = 17;
//        final int fromSz = hasParameters() ? from.size() : 0;
//        final int toSz = hasResults() ? to.size() : 0;

        int hash = assertions;
        
        if (hasParameters()) {
            int counter = 0;
            for (AbstractType t : from) {
                hash += t.hashCode() * (++counter);
            }
            hash *= firstPrime;
        } else {
            hash += firstPrime * 91;
        }

        if (hasResults()) {
            int counter = 0;
            for (AbstractType t : to) {
                hash += t.hashCode() * (++counter);
            }
            hash *= secondPrime;
        } else {
            hash += secondPrime * 91;
        }

        return hash;
    }

    public String getPrettyName() {
        return "getPrettyName(): Not implemented for FunctionalType";
    }

    public String getInternalName() {
        return "$Func" + id + assertionsToString();
    }

    public FunctionalType asLocal() {
        if (!isLocal()) {
            FunctionalType local = new FunctionalType(from, to);
            local.setAssertions(assertions+1);
            return local;
        }
        return this;
    }
}
