/*
 * COSMA -- A Virtual Machine for the Secure Cooperation of Untrusted Components.
 * Copyright (C) 2019  Roland Wismüller and Damian Ludwig, University of Siegen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package cola.ast.type;

import cola.ast.visitor.GenericVisitor;
import cola.ast.visitor.VoidVisitor;

/**
 * This class represents all primitive types.
 */
public class SimpleType extends AbstractType {
    private enum Type {
        INT("int"),             // default int
        INT8("int8"),           // one byte int
        INT16("int16"),         // two byte int
        INT32("int32"),         // four byte int
        INT64("int64"),         // eight byte int
        FLOAT("float"),         // default float
        DOUBLE("double"),       // default double
        BYTE("byte"),           // one byte
        BOOL("bool"),           // one byte
        REF("ref");             // generic reference

        private final String name;
        
        Type(String name) {
            this.name = name;
        }

        public String toString() {
            return name;
        }

        public static Type getForName(String name) {
            switch (name) {
            case "int":
                return INT;
            case "int8":
                return INT8;
            case "int16":
                return INT16;
            case "int32":
                return INT32;
            case "int64":
                return INT64;
            case "float":
                return FLOAT;
            case "double":
                return DOUBLE;
            case "byte":
                return BYTE;
            case "bool":
                return BOOL;
            case "ref":
                return REF;
            default:
                return null;
            }
        }
    }

    private final Type type;

    /**
     * Constructs a new primitive type with given kind.
     *
     * @param t the types kind
     */
    public SimpleType(Type t) {
        this.type = t;
    }

    public SimpleType(String name) {
        this(Type.getForName(name));
        ////
        assert type != null : "Unknown simple type: " + name;
        ////
    }

    public <T> void accept(VoidVisitor<T> visitor, T arg) {
        visitor.visit(this, arg);
    }

    public <T,R> R accept(GenericVisitor<T,R> visitor, T arg) {
        return visitor.visit(this, arg);
    }

    @Override
    public boolean isLocal() {
        return true;
    }

    @Override
    public boolean isFinal() {
        return true;
    }

    public String getPrettyName() {
        return type.toString();
    }

    public String getInternalName() {
        return getPrettyName();
    }

    public SimpleType asLocal() {
        return this;
    }

    @Override
    public boolean equals(Object other) {
        if (other instanceof SimpleType) {
            SimpleType t = (SimpleType) other;
            return type == t.type;
        }
        return false;
    }

    @Override
    public int hashCode() {
        return type.hashCode();
    }
}
