/*
 * COSMA -- A Virtual Machine for the Secure Cooperation of Untrusted Components.
 * Copyright (C) 2019  Roland Wismüller and Damian Ludwig, University of Siegen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package cola.ast.type;

import cola.ast.MethodNode;
import cola.ast.visitor.GenericVisitor;
import cola.ast.visitor.VoidVisitor;

import java.util.List;

/**
 * This class represents a protocol type, i.e a structural type. It is
 * a common subtype for the following language elementes:
 *
 * Structural type as a list of methods:
 *
 * {
 *   fun m1;
 *   fun m2;
 * }
 *
 *
 * Structural type derived from another type (ancestor):
 *
 * { OtherType }
 *
 *
 * Structural type bound to an extern category:
 *
 * extern {
 *   ...
 * }
 *
 *
 * Structural type as a placeholder for another typed linked at runtime:
 *
 * extern OtherType {
 *   ...
 * }
 *
 */
public class StructuralType extends AbstractType {
    private static int __nextID__ = 0;
    
    private final List<MethodNode> methods;
    private final AbstractType ancestor;
    private final String externName;
    private final boolean isExtern;
    private final int id;

    /**
     * Constructs a new protocol type with given methods.
     *
     * @param methods A list of methods
     * @param ancestor The type to derive from or null
     * @param externName Name of the linked class or null
     * @param isExtern Whether a category should be created for this type
     */
    public StructuralType(List<MethodNode> methods,
                          AbstractType ancestor,
                          String externName,
                          boolean isExtern) {
        assert !isExtern : "Extern is not supported";
        this.methods = methods;
        this.ancestor = ancestor;
        this.externName = externName;
        this.isExtern = isExtern;
        this.id = __nextID__++;
    }

    public List<MethodNode> getMethods() {
        return methods;
    }

    public boolean hasMethods() {
        return methods != null;
    }

    public AbstractType getAncestor() {
        return ancestor;
    }

    public boolean hasAncestor() {
        return ancestor != null;
    }

    public String getExternName() {
        return externName;
    }

    public boolean isExtern() {
        return isExtern;
    }

    public <T> void accept(VoidVisitor<T> visitor, T arg) {
        visitor.visit(this, arg);
    }

    public <T,R> R accept(GenericVisitor<T,R> visitor, T arg) {
        return visitor.visit(this, arg);
    }

    public String getPrettyName() {
        return "getPrettyName(): Not implemented for StructuralType";
    }

    public String getInternalName() {
        return "$Anon" + id + assertionsToString();
    }

    public StructuralType asLocal() {
        if (!isLocal()) {
            StructuralType local = new StructuralType(methods, ancestor, externName, isExtern);
            local.setAssertions(assertions+1);
            return local;
        }
        return this;
    }
}
