/*
 * COSMA -- A Virtual Machine for the Secure Cooperation of Untrusted Components.
 * Copyright (C) 2019  Roland Wismüller and Damian Ludwig, University of Siegen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package cola.ast.visitor;

import cola.ast.*;
import cola.ast.exp.*;
import cola.ast.stm.*;
import cola.ast.type.*;

public class GenericVisitorAdapter<T,R> extends GenericVisitor<T,R> {
    public R visit(AttributeNode node, T arg) {
        return node.getType().accept(this, arg);
    }

    public R visit(ComponentNode node, T arg) {
        if (node.getTypedefs() != null) {
            for (TypedefNode td : node.getTypedefs()) {
                R res = td.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        if (node.getClasses() != null) {
            for (ClassNode cls : node.getClasses()) {
                R res = cls.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        return null;
    }

    public R visit(ClassNode node, T arg) {
        if (node.getAttributeNodes() != null) {
            for (AttributeNode attr : node.getAttributeNodes()) {
                R res = attr.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        if (node.getMethods() != null) {
            for (MethodNode meth : node.getMethods()) {
                R res = meth.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        return null;
    }

    public R visit(MethodNode node, T arg) {
        if (node.getParameters() != null) {
            for (IdentifierWithType tid : node.getParameters()) {
                R res = tid.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        if (node.getResultTypes() != null) {
            for (AbstractType t : node.getResultTypes()) {
                R res = t.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        if (node.getBody() != null) {
            R res = node.getBody().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(TypedefNode node, T arg) {
        return node.getDefinition().accept(this, arg);
    }

    public R visit(IdentifierWithType node, T arg) {
        return node.getType().accept(this, arg);
    }

    public R visit(ArrayType node, T arg) {
        return node.getElementType().accept(this, arg);
    }

    public R visit(FunctionalType node, T arg) {
        if (node.getParameterTypes() != null) {
            for (AbstractType t : node.getParameterTypes()) {
                R res = t.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        if (node.getResultTypes() != null) {
            for (AbstractType t : node.getResultTypes()) {
                R res = t.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        return null;
    }

    public R visit(TypeIdentifier node, T arg) {
        return null;
    }

    public R visit(SimpleType node, T arg) {
        return null;
    }

    public R visit(StructuralType node, T arg) {
        if (node.getMethods() != null) {
            for (MethodNode meth : node.getMethods()) {
                R res = meth.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        if (node.getAncestor() != null) {
            R res = node.getAncestor().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(BlockStm node, T arg) {
        if (node.getStms() != null) {
            for (Stm stm : node.getStms()) {
                R res = stm.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        return null;
    }

    public R visit(ExpStm node, T arg) {
        return node.getExp().accept(this, arg);
    }

    public R visit(ReturnStm node, T arg) {
        if (node.getResults() != null) {
            for (Exp e : node.getResults()) {
                R res = e.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        return null;
    }

    public R visit(AssignStm node, T arg) {
        {
            R res = node.getLHS().accept(this, arg);
            if (res != null)
                return null;
        }
        {
            R res = node.getRHS().accept(this, arg);
            if (res != null)
                return null;
        }
        return null;
    }

    public R visit(VarStm node, T arg) {
        if (node.getType() != null) {
            R res = node.getType().accept(this, arg);
            if (res != null)
                return res;
        }
        if (node.getInitializer() !=  null) {
            R res = node.getInitializer().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(IfStm node, T arg) {
        {
            R res = node.getCondition().accept(this, arg);
            if (res != null)
                return res;
        }
        {
            R res = node.getThenStm().accept(this, arg);
            if (res != null)
                return res;
        }
        if (node.getElseStm() != null) {
            R res = node.getElseStm().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(WhileStm node, T arg) {
        if (node.getCondition() != null) {
            R res = node.getCondition().accept(this, arg);
            if (res != null)
                return res;
        }
        if (node.getBody() != null) {
            R res = node.getBody().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(ForStm node, T arg) {
        if (node.getInitializer() != null) {
            R res = node.getInitializer().accept(this, arg);
            if (res != null)
                return res;
        }
        if (node.getCondition() != null) {
            R res = node.getCondition().accept(this, arg);
            if (res != null)
                return res;
        }
        if (node.getAfterthought() != null) {
            R res = node.getAfterthought().accept(this, arg);
            if (res != null)
                return res;
        }
        if (node.getBody() != null) {
            R res = node.getBody().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(BreakStm node, T arg) {
        return null;
    }

    public R visit(ContinueStm node, T arg) {
        return null;
    }

    public R visit(InvokeStm node, T arg) {
        {
            R res = node.getTarget().accept(this, arg);
            if (res != null)
                return res;
        }
        {
            R res = node.getMethod().accept(this, arg);
            if (res != null)
                return res;
        }
        if (node.getArgs() != null) {
            R res = node.getArgs().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(AsmStm node, T arg) {
        if (node.getArgs() != null) {
            for (Exp e : node.getArgs()) {
                R res = e.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        return null;
    }

    public R visit(TryCatchStm node, T arg) {
        {
            R res = node.getTryStm().accept(this, arg);
            if (res != null)
                return res;
        }
        {
            R res = node.getCatchStm().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(SwitchStm node, T arg) {
        {
            R res = node.getSwitchExp().accept(this, arg);
            if (res != null)
                return res;
        }
        for (Exp e : node.getCases().keySet()) {
            {
                R res = e.accept(this, arg);
                if (res != null)
                    return res;
            }
            {
                Stm stm = node.getCases().get(e);
                if (stm != null) {
                    R res = stm.accept(this, arg);
                    if (res != null)
                        return res;
                }
            }
        }
        if (node.getDefaultCase() != null) {
            R res = node.getDefaultCase().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(IdExp node, T arg) {
        return null;
    }

    public R visit(LiteralExp node, T arg) {
        return null;
    }
    
    public R visit(UnaryExp node, T arg) {
        return node.getExp().accept(this, arg);
    }
    
    public R visit(EnclosedExp node, T arg) {
        return node.getExp().accept(this, arg);
    }

    public R visit(BinaryExp node, T arg) {
        {
            R res = node.getLHS().accept(this, arg);
            if (res != null)
                return res;
        }
        {
            R res = node.getRHS().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(ConditionalExp node, T arg) {
        {
            R res = node.getLHS().accept(this, arg);
            if (res != null)
                return res;
        }
        {
            R res = node.getRHS().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(CallExp node, T arg) {
        if (node.getTarget() != null) {
            R res = node.getTarget().accept(this, arg);
            if (res != null)
                return res;
        }
        if (node.getArgs() != null) {
            for (Exp e : node.getArgs()) {
                R res = e.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        return null;
    }

    public R visit(NewExp node, T arg) {
        if (node.getArgs() != null) {
            for (Exp e : node.getArgs()) {
                R res = e.accept(this, arg);
                if (res != null)
                    return res;
            }
        }
        return null;
    }

    public R visit(NewArrayExp node, T arg) {
        {
            R res = node.getType().accept(this, arg);
            if (res != null)
                return res;
        }
        for (Exp e : node.getSizes()) {
            R res = e.accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(FieldAccessExp node, T arg) {
        if (node.getTarget() != null) {
            R res = node.getTarget().accept(this, arg);
            if (res != null)
                return res;
        }
        {
            R res = node.getField().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(TypeVerdictExp node, T arg) {
        {
            R res = node.getExp().accept(this, arg);
            if (res != null)
                return res;
        }
        {
            R res = node.getType().accept(this, arg);
            if (res != null)
                return res;
        }
        return null;
    }

    public R visit(AnonFunctionExp node, T arg) {
        return null;
    }

    public R visit(IncreaseOrDecreaseExp node, T arg) {
        return null;
    }

    public R visit(ThisExp node, T arg) {
        return null;
    }
}
