/*
 * COSMA -- A Virtual Machine for the Secure Cooperation of Untrusted Components.
 * Copyright (C) 2019  Roland Wismüller and Damian Ludwig, University of Siegen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package cola.ast.visitor;

import cola.ast.*;
import cola.ast.exp.*;
import cola.ast.stm.*;
import cola.ast.type.*;

import java.util.List;

public class VoidVisitorAdapter<T> extends VoidVisitor<T> {
    public void visit(AttributeNode node, T arg) {
        node.getType().accept(this, arg);
    }
    
    public void visit(ComponentNode node, T arg) {
        if (node.getTypedefs() != null)
            node.getTypedefs().forEach(v -> v.accept(this, arg));
        if (node.getClasses() != null)
            node.getClasses().forEach(v -> v.accept(this, arg));
    }
    
    public void visit(ClassNode node, T arg) {
        if (node.getAttributeNodes() != null)
            node.getAttributeNodes().forEach(v -> v.accept(this, arg));
        if (node.getMethods() != null)
            node.getMethods().forEach(v -> v.accept(this, arg));
    }
    
    public void visit(MethodNode node, T arg) {
        if (node.getParameters() != null)
            node.getParameters().forEach(v -> v.accept(this, arg));
        if (node.getResultTypes() != null)
            node.getResultTypes().forEach(v -> v.accept(this, arg));
        if (node.getBody() != null)
            node.getBody().accept(this, arg);
    }
    
    public void visit(TypedefNode node, T arg) {
        node.getDefinition().accept(this, arg);
    }
    
    public void visit(IdentifierWithType node, T arg) {
        node.getType().accept(this, arg);
    }

    public void visit(ArrayType node, T arg) {
        node.getElementType().accept(this, arg);
    }
    
    public void visit(FunctionalType node, T arg){
        if (node.getParameterTypes() != null)
            node.getParameterTypes().forEach(v -> v.accept(this, arg));
        if (node.getResultTypes() != null)
            node.getResultTypes().forEach(v -> v.accept(this, arg));
    }
    
    public void visit(TypeIdentifier node, T arg){
    }
    
    public void visit(SimpleType node, T arg) {
    }
    
    public void visit(StructuralType node, T arg) {
        if (node.getMethods() != null)
            node.getMethods().forEach(v -> v.accept(this, arg));
        if (node.getAncestor() != null)
            node.getAncestor().accept(this, arg);
    }

    public void visit(BlockStm node, T arg) {
        visit(node.getStms(), arg);
    }

    public void visit(ExpStm node, T arg) {
        node.getExp().accept(this, arg);
    }

    public void visit(ReturnStm node, T arg) {
        if (node.getResults() != null)
            node.getResults().forEach(v -> v.accept(this, arg));
    }

    public void visit(AssignStm node, T arg) {
        node.getLHS().accept(this, arg);
        node.getRHS().accept(this, arg);
    }

    public void visit(VarStm node, T arg) {
        if (node.getType() != null) {
            node.getType().accept(this, arg);
        }
        if (node.getInitializer() !=  null) {
            node.getInitializer().accept(this, arg);
        }
    }

    public void visit(IfStm node, T arg) {
        node.getCondition().accept(this, arg);
        node.getThenStm().accept(this, arg);
        if (node.hasElse()) {
            node.getElseStm().accept(this, arg);
        }
    }

    public void visit(WhileStm node, T arg) {
        node.getCondition().accept(this, arg);
        node.getBody().accept(this, arg);
    }

    public void visit(ForStm node, T arg) {
        if (node.getInitializer() != null)
            node.getInitializer().accept(this, arg);
        node.getCondition().accept(this, arg);
        if (node.getAfterthought() != null)
            node.getAfterthought().accept(this, arg);
        node.getBody().accept(this, arg);
    }

    public void visit(BreakStm node, T arg) {
    }

    public void visit(ContinueStm node, T arg) {
    }

    public void visit(InvokeStm node, T arg) {
        node.getTarget().accept(this, arg);
        node.getMethod().accept(this, arg);
        if (node.getArgs() != null)
            node.getArgs().accept(this, arg);
    }

    public void visit(AsmStm node, T arg) {
        if (node.getArgs() != null)
            visit(node.getArgs(), arg);
    }

    public void visit(TryCatchStm node, T arg) {
        node.getTryStm().accept(this, arg);
        node.getCatchStm().accept(this, arg);
    }

    public void visit(SwitchStm node, T arg) {
        node.getSwitchExp().accept(this, arg);
        for (Exp e : node.getCases().keySet()) {
            e.accept(this, arg);
            Stm stm = node.getCases().get(e);
            if (stm != null)
                stm.accept(this, arg);
        }
        if (node.getDefaultCase() != null) {
            node.getDefaultCase().accept(this, arg);
        }
    }
    
    public void visit(IdExp node, T arg) {
    }

    public void visit(LiteralExp node, T arg) {
    }

    public void visit(UnaryExp node, T arg) {
        node.getExp().accept(this, arg);
    }

    public void visit(EnclosedExp node, T arg) {
        node.getExp().accept(this, arg);
    }

    public void visit(BinaryExp node, T arg) {
        node.getLHS().accept(this, arg);
        node.getRHS().accept(this, arg);
    }

    public void visit(ConditionalExp node, T arg) {
        node.getLHS().accept(this, arg);
        node.getRHS().accept(this, arg);
    }

    public void visit(CallExp node, T arg) {
        if (node.getTarget() != null)
            node.getTarget().accept(this, arg);
        visit(node.getArgs(), arg);
    }

    public void visit(ThisExp node, T arg) {
        // empty
    }

    public void visit(NewExp node, T arg) {
        if (node.getArgs() != null)
            visit(node.getArgs(), arg);
    }

    public void visit(NewArrayExp node, T arg) {
        node.getType().accept(this, arg);
        visit(node.getSizes(), arg);
    }

    public void visit(FieldAccessExp node, T arg) {
        if (node.getTarget() != null)
            node.getTarget().accept(this, arg);
        node.getField().accept(this, arg);
    }

    public void visit(TypeVerdictExp node, T arg) {
        node.getExp().accept(this, arg);
        node.getType().accept(this, arg);
    }

    public void visit(IncreaseOrDecreaseExp node, T arg) {
    }

    public void visit(AnonFunctionExp node, T arg) {
        visit(node.getParameters(), arg);
        visit(node.getResultTypes(), arg);
        node.getBody().accept(this, arg);
    }

    public <L extends Node> void visit(List<L> list, T arg) {
        if (list != null)
            list.forEach(v -> v.accept(this, arg));
    }
}
