package cola.compiler;

import cola.ast.AttributeNode;
import cola.ast.ClassNode;
import cola.ast.ComponentNode;
import cola.ast.MethodNode;
import cola.ast.type.TypeIdentifier;
import cola.util.MapReduce;
import cola.util.TSort;
import cosma.bytecode.ClassModifier;
import cosma.bytecode.attributeinfo.AttributeAccessFlags;
import cosma.bytecode.builder.*;
import cosma.bytecode.methodinfo.MethodAccessFlags;
import cosma.tools.logger.LogManager;
import cosma.tools.logger.Logger;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.stream.Collectors;

public class ComponentGenerator {
    private final Logger logger = LogManager.getCurrentClassLogger();
    private final Builder componentBuilder;
    private final ClassGenerator classGenerator;
    private final TypeGenerator typeGenerator;
    
    public ComponentGenerator(Builder builder) {
        this.componentBuilder = builder;
        this.classGenerator = new ClassGenerator(componentBuilder);
        this.typeGenerator = new TypeGenerator(componentBuilder);
    }

    public void generateComponent(ComponentNode node) {
        List<ClassNode> classNodes = tsort(node.getClasses());
        crawlClasses(classNodes, node.getComponentClassName());
        crawlTypes(node);
        crawlAttributes(classNodes);
        crawlMethods(classNodes);
        TypeCache.commit();
        ClassCache.commit();
        compile(classNodes);
    }

    private void compile(List<ClassNode> classNodes) {
        for (ClassNode classNode : classNodes) {
            logger.debug("Compiling class {class}", classNode.getName());
            ClassBuilder classBuilder = ClassCache.getClassForName(classNode.getName());
            for (MethodNode methodNode : classNode.getMethods()) {
                logger.debug("In class {class}, compiling method {method}", classNode.getName(),
                             methodNode.getName());
                MethodBuilder methodBuilder = classBuilder.getMethod(methodNode.getName());
                MethodBodyGenerator bodyGenerator =
                    new MethodBodyGenerator(componentBuilder, classBuilder, methodBuilder);
                methodNode.accept(bodyGenerator, null);
            }
        }
    }

    private void crawlAttributes(List<ClassNode> classNodes) {
        for (ClassNode cls : classNodes) {
            ClassBuilder classBuilder = ClassCache.getClassForName(cls.getName());
            for (ClassBuilder superclassBuilder : classBuilder.getSuperClasses()) {
                List<AttributeBuilder> attributes = superclassBuilder.getAttributes()
                    .stream()
                    .filter(a -> !a.getAccessFlags().contains(AttributeAccessFlags.Static))
                    .collect(Collectors.toList());
                attributes.forEach(classBuilder::addAttribute);
            }
            List<AttributeNode> attributes = cls.getAttributeNodes();
            if (attributes != null) {
                for (AttributeNode attr : attributes) {
                    EnumSet<AttributeAccessFlags> accessFlags = EnumSet.noneOf(AttributeAccessFlags.class);
                    if (attr.isStatic())
                        accessFlags.add(AttributeAccessFlags.Static);
                    String name = attr.getName();
                    IntoType type = TypeCache.getTypeForName(attr.getType().getInternalName());
                    classBuilder.createAttribute(type, name, accessFlags);
                }
            }
        }
    }

    private void crawlMethods(List<ClassNode> classNodes) {
        MethodGenerator methodGenerator = new MethodGenerator(componentBuilder);
        for (ClassNode cls : classNodes) {
            ClassBuilder classBuilder = ClassCache.getClassForName(cls.getName());
            for (ClassBuilder superclassBuilder : classBuilder.getSuperClasses()) {
                List<MethodBuilder> methods = superclassBuilder.getMethods()
                    .stream()
                    .filter(m -> !(m.getAccessFlags().contains(MethodAccessFlags.Private) ||
                             m.getAccessFlags().contains(MethodAccessFlags.Static) ||
                             m.getAccessFlags().contains(MethodAccessFlags.Constructor)))
                    .collect(Collectors.toList());
                methods.forEach(classBuilder::addMethod);
            }
            cls.accept(methodGenerator, classBuilder);
        }
    }
    
    private void crawlTypes(ComponentNode node) {
        new TypeIdentifier("String").accept(typeGenerator, null);
        if (node.getTypedefs() != null) {
            node.getTypedefs().forEach(v -> {
                    String typeName = v.getName();
                    IntoType definition = v.getDefinition().accept(typeGenerator, typeName);
                    /////
                    assert definition != null : "Type definition is null: " +
                        v.getDefinition().getInternalName();
                    /////
                    IntoType forTypeName = TypeCache.getTypeForName(typeName);
                    if (forTypeName == null) {
                        TypeCache.addType(typeName, definition);
                    } else {
                        /////
                        assert forTypeName instanceof ForwardDeclaredTypeBuilder :
                            "Type already exists, but is not forward declared: " + typeName;
                        /////
                        ForwardDeclaredTypeBuilder fwDeclBuilder = (ForwardDeclaredTypeBuilder) forTypeName;
                        fwDeclBuilder.resolveAsCopyOf(definition, typeName);
                    }
                });
        }
        node.accept(typeGenerator, null);
    }
    
    private void crawlClasses(List<ClassNode> classNodes, String componentName) {
        ClassNode componentClassNode = MapReduce.search(classNodes, v->v.getName().equals(componentName));
        /////
        if (componentClassNode == null) {
            logger.error("Component class {name} not found", componentName);
            System.exit(1);
        }
        /////
        
        /* The component class is somehow special and must be handled
         * separately from the other (nested) classes. */
        ClassBuilder componentClassBuilder = componentBuilder.createComponent(componentName, EnumSet.allOf(ClassModifier.class));
        ClassCache.addClass(componentName, componentClassBuilder);
        
        classNodes.forEach(v -> v.accept(classGenerator, null));
    }

    private List<ClassNode> tsort(List<ClassNode> classNodes) {
        TSort tsorter = new TSort(classNodes.size());
        List<ClassNode> withSuperClass = classNodes
            .stream()
            .filter(ClassNode::hasSuperclass)
            .collect(Collectors.toList());
        for (ClassNode cls : withSuperClass) {
            int myIndex = classNodes.indexOf(cls);
            for (String name : cls.getSuperclasses()) {
                ClassNode superClassNode = classNodes
                    .stream()
                    .filter(c -> c.getName().equals(name))
                    .findFirst()
                    .orElse(null);
                ////
                assert superClassNode != null : "Superclass not found";
                ////
                int superIndex = classNodes.indexOf(superClassNode);
                tsorter.addRelation(superIndex, myIndex);
                /////
                logger.debug("Inheritance detected: {sub} <: {super}", cls.getName(),
                             superClassNode.getName());
                /////
            }
        }
        List<Integer> compilationOrder = tsorter.sort();
        List<ClassNode> ordered = new ArrayList<>();
        for (Integer index : compilationOrder) {
            ordered.add(classNodes.get(index));
        }
        return ordered;
    }
}
