package cola.compiler;

import cosma.bytecode.builder.*;
import cosma.tools.logger.LogManager;
import cosma.tools.logger.Logger;

import java.util.AbstractMap.SimpleEntry;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Stack;

public class Environment {
    private final Logger logger = LogManager.getCurrentClassLogger();
    private final ClassBuilder classBuilder;
    private final MethodBuilder methodBuilder;
    private final Stack<SimpleEntry<String, VariableBuilder>> variables;
    private final Stack<Integer> variableStackSizes;
    private final Set<VariableBuilder> constants;
    private final Stack<Operand> operandStack;
    private final Stack<Label> nextLabelOnContinue;
    private final Stack<Label> nextLabelOnBreak;
    
    Environment(ClassBuilder classBuilder, MethodBuilder methodBuilder) {
        this.classBuilder = classBuilder;
        this.methodBuilder = methodBuilder;
        this.variables = new Stack<>();
        this.variableStackSizes = new Stack<>();
        this.constants = new HashSet<>();
        this.operandStack = new Stack<>();
        this.nextLabelOnContinue = new Stack<>();
        this.nextLabelOnBreak = new Stack<>();
    }

    // Virtual operand stack

    public void pushOperand(Operand operand) {
        operandStack.push(operand);
    }

    public Operand popOperand() {
        if (operandStack.isEmpty()) {
            logger.error("Emtpy stack detected: {where}", where());
            new RuntimeException().printStackTrace();
            System.exit(1);
        }
        return operandStack.pop();
    }

    public Operand peekOperand() {
        return operandStack.isEmpty() ? null : operandStack.peek();
    }

    // Variable scoping
    
    public void addVariable(String name, VariableBuilder variableBuilder) {
        addVariable(name, variableBuilder, false);
    }

    public void addVariable(String name, VariableBuilder variableBuilder, boolean asConst) {
        variables.push(new SimpleEntry<>(name, variableBuilder));
        logger.debug("New variable registered: {name} in {where}", name, where());
        if (asConst)
            constants.add(variableBuilder);
    }

    public boolean isConstant(VariableBuilder variableBuilder) {
        return constants.contains(variableBuilder);
    }

    public void pushBlock() {
        variableStackSizes.push(variables.size());
    }

    public void dropBlock() {
        int desiredStackSize = variableStackSizes.pop();
        int currentStackSize = variables.size();
        for (;currentStackSize>desiredStackSize;currentStackSize--)
            variables.pop();
    }

    // Loop data

    public void beginLoop(Label nextOnContinue, Label nextOnBreak) {
        nextLabelOnContinue.push(nextOnContinue);
        nextLabelOnBreak.push(nextOnBreak);
    }

    public void endLoop() {
        nextLabelOnContinue.pop();
        nextLabelOnBreak.pop();
    }

    public Label getNextLabelOnContinue() {
        if (nextLabelOnContinue.isEmpty())
            return null;
        return nextLabelOnContinue.peek();
    }

    public Label getNextLabelOnBreak() {
        if (nextLabelOnBreak.isEmpty())
            return null;
        return nextLabelOnBreak.peek();
    }

    // Name lookups
    
    public VariableBuilder lookup(String name) {
        List<VariableBuilder> candidates = variables.stream()
            .filter(v -> v.getKey().equals(name))
            .map(SimpleEntry::getValue)
            .collect(java.util.stream.Collectors.toList());
        if (candidates.isEmpty())
            return null;
        return candidates.get(candidates.size()-1);
    }

    public AttributeBuilder lookupAttribute(String name) {
        return classBuilder.getAttributes().stream()
            .filter(v -> v.getName().equals(name))
            .findFirst()
            .orElse(null);
    }

    public Operand lookupAttributeOrVariable(String name) {
        AttributeBuilder attr = lookupAttribute(name);
        VariableBuilder var = lookup(name);
        if (var != null)
            return new Operand(var.toOperand(), var.getType(), var.getName());
        if (attr != null)
            return new Operand(attr.toOperand(), attr.getType(), attr.getName());
        return null;
    }

    // Helpers
    
    private String where() {
        return String.format("%s::%s", classBuilder.getName(), methodBuilder.getName());
    }
}
