/*
 * COSMA -- A Virtual Machine for the Secure Cooperation of Untrusted Components.
 * Copyright (C) 2019  Roland Wismüller and Damian Ludwig, University of Siegen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package cola.compiler;

import cola.util.StringEscapeUtils;

import cosma.bytecode.builder.*;
import cosma.tools.logger.LogManager;
import cosma.tools.logger.Logger;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class InstructionEmitter {
    private static final Pattern PatternList = Pattern.compile("\\[(?:(?:%(\\d+))\\s?)*]");
    private static final Pattern PatternPosition = Pattern.compile("(?:%(\\d+))");
    private static final Pattern PatternInt = Pattern.compile("\\d+");
    private static final Pattern PatternFloat = Pattern.compile("\\d+\\.\\d+");
    private static final Pattern PatternString = Pattern.compile("\"([^']+)\"");
    
    private static Logger logger = LogManager.getCurrentClassLogger();

    private final BytecodeBuilder bodyBuilder;
    
    InstructionEmitter(BytecodeBuilder bodyBuilder) {
        this.bodyBuilder = bodyBuilder;
    }

    public void emit(String[] instructions, Object... args) {
        for (String instruction : instructions)
            emit(instruction, args);
    }

    public void emit(String instruction, Object... args) {
        instruction = instruction.replaceAll("\\s+", " ");
        String[] splitted = instruction.split(" ");

        String opcode = splitted[0];

        switch (opcode) {
        case "call":
            emitCall(instruction, args);
            break;
        case "calli":
            emitCallImmediate(instruction, args);
            break;
        case "calls":
            emitCallStatic(instruction, args);
            break;
        case "jmp":
            emitJump(instruction, args);
            break;
        case "jz":
            emitCJump(instruction, args, false);
            break;
        case "jnz":
            emitCJump(instruction, args, true);
            break;
        case "ldi":
            emitLoad(instruction, args);
            break;
        case "mov":
            emitMove(instruction, args);
            break;
        case "add":
            emitOp(instruction, args, 0);
            break;
        case "sub":
            emitOp(instruction, args, 1);
            break;
        case "mul":
            emitOp(instruction, args, 2);
            break;
        case "div":
            emitOp(instruction, args, 3);
            break;
        case "mod":
            emitOp(instruction, args, 4);
            break;
        case "eq":
            emitTest(instruction, args, 0);
            break;
        case "neq":
            emitTest(instruction, args, 1);
            break;
        case "lt":
            emitTest(instruction, args, 2);
            break;
        case "geq":
            emitTest(instruction, args, 3);
            break;
        case "gt":
            emitTest(instruction, args, 4);
            break;
        case "leq":
            emitTest(instruction, args, 5);
            break;
        case "ret":
            emitReturn(instruction, args);
            break;
        case "tsl":
            emitTsl(instruction, args);
            break;
        case "new":
            emitNew(instruction, args);
            break;
        case "na":
            emitNewA(instruction, args);
            break;
        case "ia":
            emitInitA(instruction, args);
            break;
        case "chkt":
            emitInstanceOf(instruction, args);
            break;
        case "trnc":
            emitTrunc(instruction, args);
            break;
        case "print":
            emitPrint(instruction, args);
            break;
        default:
            logger.error("(CASM) Unknown opcode: {instruction}", instruction);
            System.exit(1);
            break;
        }
    }

    private void emitReturn(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        if ((splitted != null) && (splitted.length > 0)) {
            List<Operand> results = getParameter(splitted[0], args);
            bodyBuilder.emitReturn(read(results));
        } else {
            bodyBuilder.emitReturn();
        }
    }

    private void emitCall(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        Operand scope = getParameter(splitted[0], args);
        MethodBuilder method = getParameter(splitted[1], args);
        List<Operand> parameters = getParameter(splitted[2], args);
        List<IntoOperand> results = getParameter(splitted[3], args);
        bodyBuilder.emitCallLate(read(scope), method, read(parameters), results);
    }

    private void emitCallImmediate(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        Operand scope = getParameter(splitted[0], args);
        MethodBuilder method = getParameter(splitted[1], args);
        List<Operand> parameters = getParameter(splitted[2], args);
        List<IntoOperand> results = getParameter(splitted[3], args);
        ClassBuilder targetClass = getParameter(splitted[4], args);
        bodyBuilder.emitCallImmediate(read(scope), method, targetClass, read(parameters), results);
    }

    private void emitCallStatic(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        MethodBuilder method = getParameter(splitted[0], args);
        List<Operand> parameters = getParameter(splitted[1], args);
        List<IntoOperand> results = getParameter(splitted[2], args);
        ClassBuilder targetClass = getParameter(splitted[3], args);
        bodyBuilder.emitCallStatic(method, targetClass, read(parameters), results);
    }

    private void emitMove(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        Operand dst = getParameter(splitted[0], args);
        Operand src = getParameter(splitted[1], args);
        if (dst.isArrayAccess()) {
            MethodBuilder write = dst.getType().getMethod("write");
            List<IntoOperand> parameters = read(dst.getIndices());
            parameters.add(read(src));
            List<IntoOperand> results = new ArrayList<>();
            bodyBuilder.emitCallLate(dst, write, parameters, results);
        } else {
            bodyBuilder.emitMove(dst, read(src));
        }
    }

    private void emitOp(String instruction, Object[] args, int operation) {
        String[] splitted = tokenize(instruction);

        IntoOperand dst = getParameter(splitted[0], args);
        Operand lhs = getParameter(splitted[1], args);
        Operand rhs = getParameter(splitted[2], args);

        switch (operation) {
        case 0:
            bodyBuilder.emitOpPlus(dst, read(lhs), read(rhs));
            break;
        case 1:
            bodyBuilder.emitOpMinus(dst, read(lhs), read(rhs));
            break;
        case 2:
            bodyBuilder.emitOpMult(dst, read(lhs), read(rhs));
            break;
        case 3:
            bodyBuilder.emitOpDiv(dst, read(lhs), read(rhs));
            break;
        case 4:
            bodyBuilder.emitOpMod(dst, read(lhs), read(rhs));
            break;
        default:
            throw new RuntimeException("Wrong opcode");
        }
    }

    private void emitTest(String instruction, Object[] args, int operation) {
        String[] splitted = tokenize(instruction);

        IntoOperand dst = getParameter(splitted[0], args);
        Operand lhs = getParameter(splitted[1], args);
        Operand rhs = getParameter(splitted[2], args);

        switch (operation) {
        case 0:
            bodyBuilder.emitTestEq(dst, read(lhs), read(rhs));
            break;
        case 1:
            bodyBuilder.emitTestNeq(dst, read(lhs), read(rhs));
            break;
        case 2:
            bodyBuilder.emitTestLt(dst, read(lhs), read(rhs));
            break;
        case 3:
            bodyBuilder.emitTestGeq(dst, read(lhs), read(rhs));
            break;
        case 4:
            bodyBuilder.emitTestGt(dst, read(lhs), read(rhs));
            break;
        case 5:
            bodyBuilder.emitTestLeq(dst, read(lhs), read(rhs));
            break;
        default:
            throw new RuntimeException("Wrong opcode");
        }
    }

    private void emitJump(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        Label target = getParameter(splitted[0], args);
        bodyBuilder.emitJump(target);
    }

    private void emitLoad(String instruction, Object[] args) {
        emitMove(instruction, args);
    }

    private void emitCJump(String instruction, Object[] args, boolean ifNotZero) {
        String[] splitted = tokenize(instruction);
        Operand condition = getParameter(splitted[0], args);
        Label target = getParameter(splitted[1], args);
        bodyBuilder.emitConditionalJump(target, read(condition), ifNotZero);
    }

    private void emitTsl(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        IntoOperand copy = getParameter(splitted[1], args);
        IntoOperand lock = getParameter(splitted[0], args);
        bodyBuilder.emitTsl(copy, lock);
    }

    private void emitNew(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        ClassBuilder cls = getParameter(splitted[1], args);
        IntoOperand dst = getParameter(splitted[0], args);
        bodyBuilder.emitNew(false, dst, cls);
    }

    private void emitNewA(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        IntoOperand array = getParameter(splitted[0], args);
        IntoType type = getParameter(splitted[1], args);
        List<Operand> sizes = getParameter(splitted[2], args);
        bodyBuilder.emitNewA(array, type, read(sizes));
    }

    private void emitInitA(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        IntoOperand array = getParameter(splitted[0], args);
        int[] values = getParameter(splitted[1], args);
        bodyBuilder.emitInitA(array, values);
    }

    private void emitInstanceOf(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        IntoOperand dst = getParameter(splitted[0], args);
        Operand ref = getParameter(splitted[1], args);
        IntoType type = getParameter(splitted[2], args);
        bodyBuilder.emitInstanceof(dst, read(ref), type);
    }

    private void emitTrunc(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        IntoOperand dst = getParameter(splitted[0], args);
        Operand src = getParameter(splitted[1], args);
        bodyBuilder.emitTrunc(dst, read(src));
    }

    private void emitPrint(String instruction, Object[] args) {
        String[] splitted = tokenize(instruction);
        String msg = getParameter(splitted[0], args);
        IntoOperand op = null;
        if (splitted.length == 2) {
            op = getParameter(splitted[1], args);
        }
        bodyBuilder.emitPrint(msg, op);
    }

    // private helpers

    private static int __nextId__ = 0;

    private IntoOperand read(Operand operand) {
        if (operand.isArrayAccess()) {
            MethodBuilder read = operand.getType().getMethod("read");
            IntoType readType = operand.getReadType();
            VariableBuilder tmp = bodyBuilder.createVariable(readType, "$read"+__nextId__++);
            List<IntoOperand> results = new ArrayList<>();
            Operand result = new Operand(tmp.toOperand(), readType, null);
            results.add(result);
            bodyBuilder.emitCallLate(operand, read, read(operand.getIndices()), results);
            return result;
        } else {
            return operand;
        }
    }

    private List<IntoOperand> read(List<Operand> operands) {
        List<IntoOperand> res = new ArrayList<>();
        for (Operand o : operands) {
            res.add(read(o));
        }
        return res;
    }

    // public helpers

    @SuppressWarnings("unchecked")
    public <R> R getParameter(String f, Object[] args) {
        if (isPosition(f)) {
            int pos = getPosition(f);
            return (R) args[pos];
        }

        if (isList(f)) {
            List<Object> list = new ArrayList<>();
            String[] tokens = tokenizeList(f);
            for (String tok : tokens) {
                int pos = getPosition(tok);
                list.add(args[pos]);
            }
            return (R) list;
        }

        if (isImmediate(f)) {
            return (R) getImmediate(f);
        }
                
        logger.warning("Cannot be matched as parameter: {f}", f);
        return null;
    }

    public String[] tokenize(String s) {
        List<String> tokens = new ArrayList<>();

        Pattern p = Pattern.compile(PatternPosition.pattern() + "|" +
                                    PatternList.pattern() + "|" +
                                    PatternInt.pattern() + "|" +
                                    PatternFloat.pattern() + "|" +
                                    PatternString.pattern());
                
        Matcher m = p.matcher(s);

        while (m.find()) {
            tokens.add(m.group());
        }

        return tokens.toArray(new String[0]);
    }

    public String[] tokenizeList(String s) {
        List<String> tokens = new ArrayList<>();
                
        Pattern p = Pattern.compile(PatternPosition.pattern());
        Matcher m = p.matcher(s);
                
        while (m.find()) {
            tokens.add(m.group());
        }
                
        return tokens.toArray(new String[0]);
    }

    public boolean isPosition(String s) {
        return s.matches(PatternPosition.pattern());
    }

    public boolean isList(String s) {
        return s.matches(PatternList.pattern());
    }

    public boolean isImmediate(String s) {
        return s.matches(PatternInt.pattern()) ||
            s.matches(PatternFloat.pattern()) ||
            s.matches(PatternString.pattern());
    }

    public int getPosition(String s) {
        if (isPosition(s))
            return Integer.parseInt(s.substring(1));
        return -1;
    }

    public Object getImmediate(String s) {
        if (!isImmediate(s))
            return null;

        Matcher intMatcher = PatternInt.matcher(s);
        if (intMatcher.matches()) {
            return Integer.parseInt(intMatcher.group(0));
        }
                
        Matcher fltMatcher = PatternFloat.matcher(s);
        if (fltMatcher.matches()) {
            return Float.parseFloat(fltMatcher.group(0));
        }

        Matcher strMatcher = PatternString.matcher(s);
        if (strMatcher.matches()) {
            return StringEscapeUtils.unescapeJava(strMatcher.group(1));
        }
                
        return null;
    }

    public Boolean getBoolean(String s) {
        return s.equals("true");
    }
}
