/*
 * COSMA -- A Virtual Machine for the Secure Cooperation of Untrusted Components.
 * Copyright (C) 2019  Roland Wismüller and Damian Ludwig, University of Siegen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package cola.compiler;

import cola.ast.Node;
import cola.ast.exp.*;
import cola.ast.stm.*;
import cola.ast.type.ArrayType;
import cola.ast.visitor.VoidVisitorAdapter;
import cola.util.MapReduce;
import cola.util.StringEscapeUtils;
import cosma.bytecode.ClassModifier;
import cosma.bytecode.TypeProtocol;
import cosma.bytecode.attributeinfo.AttributeAccessFlags;
import cosma.bytecode.builder.*;
import cosma.bytecode.methodinfo.MethodAccessFlags;
import cosma.tools.logger.LogManager;
import cosma.tools.logger.Logger;

import java.util.*;

public class MethodBodyGenerator extends VoidVisitorAdapter<Object> {
    private final Logger logger = LogManager.getCurrentClassLogger();
    private final Builder componentBuilder;
    private final ClassBuilder classBuilder;
    private final MethodBuilder methodBuilder;
    private final BytecodeBuilder bodyBuilder;
    private final InstructionEmitter emitter;
    private final Environment environment;
    private MethodBodyGenerator parent;
    private int __nextTempID__ = 0;
    private int __nextClosureID__ = 0;
    
    MethodBodyGenerator(Builder componentBuilder,
                        ClassBuilder classBuilder,
                        MethodBuilder methodBuilder) {
        this.componentBuilder = componentBuilder;
        this.classBuilder = classBuilder;
        this.methodBuilder = methodBuilder;
        this.bodyBuilder = methodBuilder.createBody();
        this.emitter = new InstructionEmitter(bodyBuilder);
        this.environment = new Environment(classBuilder, methodBuilder);
        methodBuilder.getArguments().forEach(v -> environment.addVariable(v.getName(), v));
    }

    public MethodBodyGenerator withParent(MethodBodyGenerator parent) {
        this.parent = parent;
        return this;
    }
    
    @Override
    public void visit(BlockStm node, Object arg) {
        environment.pushBlock();
        super.visit(node, arg);
        environment.dropBlock();
    }

    @Override
    public void visit(AssignStm node, Object arg) {
        node.getLHS().accept(this, arg);
        Operand lhs = environment.popOperand();
        node.getRHS().accept(this, arg);
        Operand rhs = environment.popOperand();

        if (needsTrunc(lhs.getReadType(), rhs.getReadType(), true)) {
            Operand tmp = getNewTemp(lhs.getReadType());
            emitter.emit("trnc %0 %1", tmp, rhs);
            emitter.emit("mov %0 %1", lhs, tmp);
        } else {
            emitter.emit("mov %0 %1", lhs, rhs);
        }
    }

    @Override
    public void visit(VarStm node, Object arg) {
        String variableName = node.getName();
        if (node.getType() != null) {
            IntoType type = TypeCache.getTypeForName(node.getType().getInternalName());
            VariableBuilder variable = bodyBuilder.createVariable(type, variableName);
            environment.addVariable(variableName, variable, node.isConst());
            if (node.getInitializer() != null) {
                Operand dst = new Operand(variable.toOperand(), type, variableName);
                node.getInitializer().accept(this, arg);
                Operand src = environment.popOperand();
                if (needsTrunc(dst.getReadType(), src.getReadType(), true))
                    emitter.emit("trnc %0 %1", dst, src);
                else
                    emitter.emit("mov %0 %1", dst, src);
            } else if (node.isConst()) {
                /////
                logger.warning("Constand is not initialized: {node}", node.toString());
                /////
            }
        } else if (node.getInitializer() != null) {
            node.getInitializer().accept(this, arg);
            Operand src = environment.popOperand();
            VariableBuilder variable = bodyBuilder.createVariable(src.getReadType(), variableName);
            environment.addVariable(variableName, variable, node.isConst());
            Operand dst = new Operand(variable.toOperand(), variable.getType(), variableName);
            emitter.emit("mov %0 %1", dst, src);
        } else {
            /////
            logger.error("Variable has no type and no initializer: {node}", node.toString());
            System.exit(1);
            /////
        }
    }

    @Override
    public void visit(ReturnStm node, Object arg) {
        if (node.hasOperands()) {
            int numOperands = node.getOperands().size();
            /////
            if (methodBuilder.getNumReturnTypes() != numOperands) {
                logger.error("Formal and actual parameters differ in length: {node}", node.toString());
                System.exit(1);
            }
            /////
            visitReverse(node.getOperands(), arg);
            List<IntoOperand> operands = new ArrayList<>();
            for (int i=0;i<numOperands;i++) {
                operands.add(environment.popOperand());
            }
            emitter.emit("ret %0", operands);
        } else {
            emitter.emit("ret");
        }
    }

    @Override
    public void visit(IfStm node, Object arg) {
        node.getCondition().accept(this, arg);
        Operand condition = environment.popOperand();
        Label lblAfter = bodyBuilder.createLabel();
        Label lblElse = node.hasElse() ? bodyBuilder.createLabel() : null;
        if (lblElse != null)
            emitter.emit("jz %0 %1", condition, lblElse);
        else
            emitter.emit("jz %0 %1", condition, lblAfter);
        node.getThenStm().accept(this, arg);
        if (lblElse != null) {
            emitter.emit("jmp %0", lblAfter);
            bodyBuilder.markLabel(lblElse);
            node.getElseStm().accept(this, arg);
        }
        bodyBuilder.markLabel(lblAfter);
    }

    @Override
    public void visit(WhileStm node, Object arg) {
        Label whileBody = bodyBuilder.createLabel();
        Label afterWhile = bodyBuilder.createLabel();
        bodyBuilder.markLabel(whileBody);
        environment.beginLoop(whileBody, afterWhile);
        node.getCondition().accept(this, arg);
        Operand condition = environment.popOperand();
        emitter.emit("jz %0 %1", condition, afterWhile);
        node.getBody().accept(this, arg);
        emitter.emit("jmp %0", whileBody);
        bodyBuilder.markLabel(afterWhile);
        environment.endLoop();
    }

    @Override
    public void visit(ForStm node, Object arg) {
        Label forBody = bodyBuilder.createLabel();
        Label afterFor = bodyBuilder.createLabel();
        if (node.getInitializer() != null)
            node.getInitializer().accept(this, arg);
        environment.beginLoop(forBody, afterFor);
        bodyBuilder.markLabel(forBody);
        if (node.getCondition() != null) {
            node.getCondition().accept(this, arg);
            Operand condition = environment.popOperand();
            emitter.emit("jz %0 %1", condition, afterFor);
        }
        node.getBody().accept(this, arg);
        if (node.getAfterthought() != null)
            node.getAfterthought().accept(this, arg);
        emitter.emit("jmp %0", forBody);
        bodyBuilder.markLabel(afterFor);
        environment.endLoop();
    }

    @Override
    public void visit(ExpStm node, Object arg) {
        node.getExp().accept(this, arg);
    }

    @Override
    public void visit(BreakStm node, Object arg) {
        Label target = environment.getNextLabelOnBreak();
        if (target == null) {
            /////
            logger.warning("No matching loop: {node}", node.toString());
            /////
        } else {
            emitter.emit("jmp %0", target);
        }
    }

    @Override
    public void visit(ContinueStm node, Object arg) {
        Label target = environment.getNextLabelOnContinue();
        if (target == null) {
            /////
            logger.warning("No matching loop: {node}", node.toString());
            /////
        } else {
            emitter.emit("jmp %0", target);
        }
    }

    @Override
    public void visit(AsmStm node, Object arg) {
        String[] code = node.getCode().split("\n");
        Object[] args = new Object[node.getArgs().size()];
        for (String line : code) {
            String[] splitted = line.split(" ", 2);
            String opcode = splitted[0];
            String[] parameters = emitter.tokenize(splitted[1]);

            switch (opcode) {
            case "call": {
                int scopePosition = emitter.getPosition(parameters[0]);
                Operand scope = (Operand) args[scopePosition];
                if (scope == null) {
                    node.getArgs().get(scopePosition).accept(this, arg);
                    scope = environment.popOperand();
                    args[scopePosition] = scope;
                }
                int methodPosition = emitter.getPosition(parameters[1]);
                MethodBuilder method = (MethodBuilder) args[methodPosition];
                if (method == null) {
                    String methodName = ((LiteralExp) node.getArgs().get(methodPosition)).getStringValue();
                    // IntoType scopeType = scope.toOperand().getReadType(classBuilder.getClassType(), null);
                    IntoType scopeType = scope.getType();
                    method = scopeType.getMethod(methodName);
                    args[methodPosition] = method;
                }
                String[] callArgs = emitter.tokenizeList(parameters[2]);
                for (String callArg : callArgs) {
                    int pos = emitter.getPosition(callArg);
                    if (args[pos] == null) {
                        node.getArgs().get(pos).accept(this, arg);
                        args[pos] = environment.popOperand();
                    }
                }
                String[] callRes = emitter.tokenizeList(parameters[3]);
                for (String res : callRes) {
                    int pos = emitter.getPosition(res);
                    if (args[pos] == null) {
                        node.getArgs().get(pos).accept(this, arg);
                        args[pos] = environment.popOperand();
                    }
                }
                break;
            }
            case "ldi": {
                /* destination */
                int dstPosition = emitter.getPosition(parameters[0]);
                if (args[dstPosition] == null) {
                    node.getArgs().get(dstPosition).accept(this, arg);
                    args[dstPosition] = environment.popOperand();
                }
                // no need to process immediates
                break;
            }
            case "new":
                break;
            case "mov":
            case "tsl": {
                /* destination */
                int dstPosition = emitter.getPosition(parameters[0]);
                if (args[dstPosition] == null) {
                    node.getArgs().get(dstPosition).accept(this, arg);
                    args[dstPosition] = environment.popOperand();
                }
                /* source */
                int srcPosition = emitter.getPosition(parameters[1]);
                if (args[srcPosition] == null) {
                    node.getArgs().get(srcPosition).accept(this, arg);
                    args[srcPosition] = environment.popOperand();
                }
                break;
            }
            case "add":
            case "sub":
            case "mul":
            case "div":
            case "mod":
            case "eq":
            case "neq":
            case "lt":
            case "geq":
            case "gt":
            case "leq": {
                /* destination */
                int dstPosition = emitter.getPosition(parameters[0]);
                if (args[dstPosition] == null) {
                    node.getArgs().get(dstPosition).accept(this, arg);
                    args[dstPosition] = environment.popOperand();
                }
                /* left hand side */
                int lhsPosition = emitter.getPosition(parameters[1]);
                if (args[lhsPosition] == null) {
                    node.getArgs().get(lhsPosition).accept(this, arg);
                    args[lhsPosition] = environment.popOperand();
                }
                /* right hand side */
                int rhsPosition = emitter.getPosition(parameters[2]);
                if (args[rhsPosition] == null) {
                    node.getArgs().get(rhsPosition).accept(this, arg);
                    args[rhsPosition] = environment.popOperand();
                }
                break;
            }
            case "ret": {
                String[] results = emitter.tokenizeList(parameters[0]);
                for (String res : results) {
                    int pos = emitter.getPosition(res);
                    if (args[pos] == null) {
                        node.getArgs().get(pos).accept(this, arg);
                        args[pos] = environment.popOperand();
                    }
                }
                break;
            }
            case "print": {
                if (parameters.length > 0) {
                    int pos = emitter.getPosition(parameters[1]);
                    if (args[pos] == null) {
                        node.getArgs().get(pos).accept(this, arg);
                        args[pos] = environment.popOperand();
                    }
                }
            }
            }
        }
        emitter.emit(code, args);
    }

    @Override
    public void visit(SwitchStm node, Object arg) {
        Label afterSwitch = bodyBuilder.createLabel();
        Map<Exp, Label> caseLabels = new HashMap<>();
        node.getSwitchExp().accept(this, arg);
        Operand switchOperand = environment.popOperand();
        Operand testResult = getNewTemp(PredefinedType.Int8);
        for (Exp e : node.getCases().keySet()) {
            e.accept(this, arg);
            Operand caseOperand = environment.popOperand();
            Label caseLabel = bodyBuilder.createLabel();
            String[] checkAndJump = {
                "eq %3 %0 %1",
                "jnz %3 %2"
            };
            emitter.emit(checkAndJump, caseOperand, switchOperand, caseLabel, testResult);
            caseLabels.put(e, caseLabel);
        }
        if (node.getDefaultCase() != null) {
            node.getDefaultCase().accept(this, arg);
        }
        emitter.emit("jmp %0", afterSwitch);
        environment.beginLoop(afterSwitch, afterSwitch);
        for (Exp e : node.getCases().keySet()) {
            Stm caseStm = node.getCases().get(e);
            Label caseLabel = caseLabels.get(e);
            bodyBuilder.markLabel(caseLabel);
            if (caseStm != null)
                caseStm.accept(this, arg);
        }
        bodyBuilder.markLabel(afterSwitch);
        environment.endLoop();
    }

    @Override
    public void visit(TryCatchStm node, Object arg) {
        bodyBuilder.beginTryBlock();
        node.getTryStm().accept(this, arg);
        bodyBuilder.beginCatchBlock();
        node.getCatchStm().accept(this, arg);
        bodyBuilder.endCatchBlock();
    }

    // Expressions

    @Override
    public void visit(LiteralExp node, Object arg) {
        if (node.isNull()) {
            Operand val = new Operand(OperandBuilder.nullOperand(componentBuilder), null, null);
            environment.pushOperand(val);
        } else if (node.isInteger()) {
            Operand val = Operand.makeConstInt32(node.getIntegerValue(), componentBuilder);
            environment.pushOperand(val);
        } else if (node.isFloat()) {
            Operand val = Operand.makeConstFloat32(node.getFloatValue(), componentBuilder);
            environment.pushOperand(val);
        } else if (node.isDouble()) {
            Operand val = Operand.makeConstFloat64(node.getFloatValue(), componentBuilder);
            environment.pushOperand(val);
        } else if (node.isBool()) {
            Operand val = Operand.makeConstBool(node.getBooleanValue(), componentBuilder);
            environment.pushOperand(val);
        } else if (node.isChar()) {
            String svalue = StringEscapeUtils.unescapeJava(node.getStringValue());
            int ascii = svalue.charAt(0);
            Operand val = Operand.makeConstChar(ascii, componentBuilder);
            environment.pushOperand(val);
        } else if (node.isString()) {
            String val = StringEscapeUtils.unescapeJava(node.getStringValue());
            int[] ascii = new int[val.length()];
            for (int i=0; i<val.length(); i++) {
                ascii[i] = val.charAt(i);
            }
            Operand string = getNewTemp(TypeCache.getTypeForName("String"));
            Operand length = Operand.makeConstInt32(ascii.length, componentBuilder);
            emitter.emit("na %0 %1 [%2]", string, PredefinedType.Int8, length);
            emitter.emit("ia %0 %1", string, ascii);
            environment.pushOperand(string);
        } else {
            /////
            logger.error("Unknown type of literal expression: {node}", node.toString());
            System.exit(1);
            /////
        }
    }

    @Override
    public void visit(UnaryExp node, Object arg) {
        node.getExp().accept(this, arg);
        Operand src = environment.popOperand();
        Operand dst = getNewTemp(src.getType());
        if (node.getOperator() == UnaryExp.Operator.MINUS) {
            emitter.emit("ldi %0 %1", dst, Operand.makeConstInt32(-1, componentBuilder));
            emitter.emit("mul %0 %0 %1", dst, src);
            environment.pushOperand(dst);
        } else if (node.getOperator() == UnaryExp.Operator.NOT) {
            emitter.emit("ldi %0 %1", dst, Operand.makeConstInt32(0, componentBuilder));
            emitter.emit("eq %0 %0 %1", dst, src);
            environment.pushOperand(dst);
        } else {
            environment.pushOperand(src);
        }
    }

    @Override
    public void visit(IdExp node, Object arg) {
        String name = node.getIdentifier();
        Operand operand = environment.lookupAttributeOrVariable(name);
        if (operand == null) {
            logger.debug("{name} is not a variable or attribute: {node}", name, node.toString());
            return;
        }
        if (node.hasIndices()) {
            int numIndices = node.getIndices().size();
            int applied = 0;
            int appliedCurrent = 0;
            int dimensions = operand.getDimensions();
            visitReverse(node.getIndices(), arg);
            for (int i=0;i<numIndices;i++) {
                //////////
                // if (dimensions == 0) {
                //     logger.error("{name} is not an array: {node}", name, node.toString());
                //     System.exit(1);
                // }
                //////////

                operand.withIndex(environment.popOperand());
                applied++;
                appliedCurrent++;

                if ((applied < numIndices) && (appliedCurrent == dimensions)) {
                    Operand tmp = getNewTemp(operand.getReadType());
                    emitter.emit("mov %0 %1", tmp, operand);
                    operand = tmp;
                    appliedCurrent = 0;
                    dimensions = operand.getDimensions();
                }
            }
            //////////
            if (appliedCurrent < dimensions) {
                logger.error("{name} has unadressed dimensions: {node}", name, node.toString());
                System.exit(1);
            }
            //////////
        }
        environment.pushOperand(operand);
    }

    @Override
    public void visit(BinaryExp node, Object arg) {
        int opcode = -1;
        String instrPfix = "";
        switch(node.getOperator()) {
        case PLUS:
            opcode = 0;
            instrPfix = "add";
            break;
        case MINUS:
            opcode = 1;
            instrPfix = "sub";
            break;
        case MULT:
            opcode = 2;
            instrPfix = "mul";
            break;
        case DIV:
            opcode = 3;
            instrPfix = "div";
            break;
        case MOD:
            opcode = 4;
            instrPfix = "mod";
            break;
        case CONS:
            opcode = 5;
            break;
        default:
            /////
            logger.error("Operation is not supported: {node}", node.toString());
            System.exit(1);
            /////
        }
        node.getLHS().accept(this, arg);
        Operand lhs = environment.popOperand();
        node.getRHS().accept(this, arg);
        Operand rhs = environment.popOperand();
        Operand dst = getNewTemp(lhs.getReadType());
        if (opcode < 5) {
            emitter.emit(String.format("%s %%0 %%1 %%2", instrPfix), dst, lhs, rhs);
            environment.pushOperand(dst);
        } else {
            Operand zeroOp = getNewTemp(PredefinedType.Int32);
            Operand leftSzOp = getNewTemp(PredefinedType.Int32);
            Operand rightSzOp = getNewTemp(PredefinedType.Int32);
            Operand resSzOp = getNewTemp(PredefinedType.Int32);

            String[] getLength_asm = {
                "ldi %0 %1",
                "call %2 %3 [%4] [%5]",
                "call %6 %7 [%4] [%8]",
                "add %9 %5 %8",
                "na %10 %11 [%9]"
            };

            emitter.emit(getLength_asm,
                         zeroOp,
                         Operand.makeConstInt32(0, componentBuilder),
                         lhs,
                         lhs.getType().getMethod("length"),
                         zeroOp,
                         leftSzOp,
                         rhs,
                         rhs.getType().getMethod("length"),
                         rightSzOp,
                         resSzOp,
                         dst,
                         PredefinedType.Int8);

            Operand walkOp = getNewTemp(PredefinedType.Int32);
            Operand testOp = getNewTemp(PredefinedType.Int8);
            Operand oneOp = zeroOp;

            String[] loopPrototype = {
                "lt %2 %0 %1",
                "jz %2 %3",
                "mov %4 %5",
                "add %0 %0 %6",
                "add %7 %7 %6",
                "jmp %8"
            };

            emitter.emit("ldi %0 %1", resSzOp, Operand.makeConstInt32(0, componentBuilder));
            emitter.emit("ldi %0 %1", walkOp, Operand.makeConstInt32(0, componentBuilder));
            emitter.emit("ldi %0 %1", oneOp, Operand.makeConstInt32(1, componentBuilder));

            Label copyLhs = bodyBuilder.createLabel();
            Label between = bodyBuilder.createLabel();
            bodyBuilder.markLabel(copyLhs);

            emitter.emit(loopPrototype,
                         walkOp,
                         leftSzOp,
                         testOp,
                         between,
                         dst.clone().withIndex(resSzOp),
                         lhs.clone().withIndex(resSzOp),
                         oneOp,
                         resSzOp,
                         copyLhs);

            bodyBuilder.markLabel(between);
            emitter.emit("ldi %0 %1", walkOp, Operand.makeConstInt32(0, componentBuilder));

            Label copyRhs = bodyBuilder.createLabel();
            Label afterCopy = bodyBuilder.createLabel();
            bodyBuilder.markLabel(copyRhs);

            emitter.emit(loopPrototype,
                         walkOp,
                         rightSzOp,
                         testOp,
                         afterCopy,
                         dst.clone().withIndex(resSzOp),
                         rhs.clone().withIndex(walkOp),
                         oneOp,
                         resSzOp,
                         copyRhs);

            bodyBuilder.markLabel(afterCopy);
            environment.pushOperand(dst);
        }
        
    }

    @Override
    public void visit(ConditionalExp node, Object arg) {
        Operand dst = getNewTemp(PredefinedType.Int8);
        String instrPfix = "";
        switch(node.getOperator()) {
        case EQ:
            instrPfix = "eq";
            break;
        case NEQ:
            instrPfix = "neq";
            break;
        case LT:
            instrPfix = "lt";
            break;
        case LEQ:
            instrPfix = "leq";
            break;
        case GT:
            instrPfix = "gt";
            break;
        case GEQ:
            instrPfix = "geq";
            break;
        }
        node.getLHS().accept(this, arg);
        Operand lhs = environment.popOperand();
        node.getRHS().accept(this, arg);
        Operand rhs = environment.popOperand();
        emitter.emit(String.format("%s %%0 %%1 %%2", instrPfix), dst, lhs, rhs);
        environment.pushOperand(dst);
    }

    @Override
    public void visit(ThisExp node, Object arg) {
        OperandBuilder thisOperand = OperandBuilder.thisOperand(componentBuilder);
        IntoType classType = classBuilder.getClassType();
        environment.pushOperand(new Operand(thisOperand, classType, "this"));
    }

    @Override
    public void visit(NewExp node, Object arg) {
        ClassBuilder cls = ClassCache.getClassForName(node.getClassName());
        IntoType clsType = cls.getClassType();
        Operand dst = getNewTemp(clsType);
        emitter.emit("new %0 %1", dst, cls);
        MethodBuilder constructor = clsType.getMethod("$init");
        if (constructor != null) {
            /////
            assert constructor.getAccessFlags().contains(MethodAccessFlags.Constructor)
                : "Found $init which is not a constructor";
            if ((classBuilder != cls) && constructor.getAccessFlags().contains(MethodAccessFlags.Private)) {
                logger.error("Constructor is private in class {name}: {node}", cls.getName(), node.toString());
                System.exit(1);
            }
            /////
            List<IntoOperand> parameters = new ArrayList<>();
            if (node.getArgs() != null) {
                /////
                if (node.getArgs().size() != constructor.getNumArguments()) {
                    logger.error("Wrong number of args for method \"$init\": found {give}, required {needed}; {node}",
                                 node.getArgs().size(), constructor.getNumArguments(), node.toString());
                    System.exit(1);
                }
                /////
                visitReverse(node.getArgs(), arg);
                for (int i=0;i<node.getArgs().size();i++) {
                    parameters.add(environment.popOperand());
                }
            }
            emitter.emit("calli %0 %1 %2 %3 %4", dst, constructor, parameters, new ArrayList<>(), cls);
        }
        environment.pushOperand(dst);
    }

    @Override
    public void visit(CallExp node, Object arg) {
        List<Operand> results = new ArrayList<>();
        Operand scope = null;
        MethodBuilder method = null;
        ClassBuilder containingClass = null;
        if (node.getTarget() != null) {
            node.getTarget().accept(this, arg);
            if (environment.peekOperand() != null) {
                scope = environment.popOperand();
                if (scope.isThis()) {
                    containingClass = classBuilder;
                    if (insideClosure()) {
                        AttributeBuilder __this__ = classBuilder.getAttribute("__this__");
                        IntoType type = __this__.getType();
                        scope = new Operand(__this__.toOperand(), type, "__this__");
                    }
                }
                method = scope.getType().getMethod(node.getMethodName());
            } else {
                /////
                if (!(node.getTarget() instanceof IdExp)) {
                    logger.error("Internal error at: {node}", node.toString());
                    System.exit(1);
                }
                /////
                String className = ((IdExp) node.getTarget()).getIdentifier();
                containingClass = ClassCache.getClassForName(className);
                method = containingClass.getMethod(node.getMethodName());
            }
        } else {
            // try for closure
            scope = environment.lookupAttributeOrVariable(node.getMethodName());
            if (scope != null) {
                method = scope.getType().getMethod("call");
            }

            // try for unscoped member method
            if (method == null) {
                new ThisExp().accept(this, arg);
                scope = environment.popOperand();
                containingClass = classBuilder;
                method = scope.getType().getMethod(node.getMethodName());
            }
        }

        /////
        if (method == null) {
            logger.error("No such method: {class}::{method} at {node}", scope.getType().getName(),
                         node.getMethodName(), node.toString());
            System.exit(1);
        }
        /////
        visitReverse(node.getArgs(), arg);
        List<Operand> args = new ArrayList<>();
        for (int i=0;i<method.getNumArguments();i++)
            args.add(environment.popOperand());
        if (method.getNumReturnTypes() > 0) {
            //results.clear();
            method.getReturnTypes().forEach(v -> results.add(getNewTemp(v)));
        }
        if (method.getAccessFlags().contains(MethodAccessFlags.Static)) {
            emitter.emit("calls %0 %1 %2 %3", method, args, results, containingClass);
        } else if (method.getAccessFlags().contains(MethodAccessFlags.Private)) {
            emitter.emit("calli %0 %1 %2 %3 %4", scope, method, args, results, containingClass);
        } else {
            emitter.emit("call %0 %1 %2 %3", scope, method, args, results);
        }
        ListIterator<Operand> it = results.listIterator(results.size());
        while (it.hasPrevious()) {
            environment.pushOperand(it.previous());
        }
    }

    @Override
    public void visit(NewArrayExp node, Object arg) {
        ArrayType colaType = node.getType();
        IntoType arrayType = TypeCache.getTypeForName(colaType.getInternalName());
        IntoType elementType = TypeCache.getTypeForName(colaType.getElementType().getInternalName());
        Operand dst = getNewTemp(arrayType);
        visitReverse(node.getSizes(), arg);
        List<IntoOperand> sizes = new ArrayList<>();
        for (int i=0;i<node.getNumDimensions();i++) {
            sizes.add(environment.popOperand());
        }
        emitter.emit("na %0 %1 %2", dst, elementType, sizes);
        environment.pushOperand(dst);
    }

    @Override
    public void visit(IncreaseOrDecreaseExp node, Object arg) {
        Operand operand = environment.lookupAttributeOrVariable(node.getIdentifier());
        /////
        if (operand == null) {
            logger.error("Cannot find symbol: {name} at {node}", node.getIdentifier(), node.toString());
            System.exit(1);
        }
        /////
        new LiteralExp(1).accept(this, arg);
        Operand one = environment.popOperand();
        if (node.isPost()) {
            Operand copy = getNewTemp(operand.getType());
            emitter.emit("mov %0 %1", copy, operand);
            if (node.isDecrease())
                emitter.emit("sub %0 %0 %1", operand, one);
            else
                emitter.emit("add %0 %0 %1", operand, one);
            environment.pushOperand(copy);
        } else {
            if (node.isDecrease())
                emitter.emit("sub %0 %0 %1", operand, one);
            else
                emitter.emit("add %0 %0 %1", operand, one);
            environment.pushOperand(operand);
        }
    }

    @Override
    public void visit(TypeVerdictExp node, Object arg) {
        node.getExp().accept(this, arg);
        Operand toCheck = environment.popOperand();
        IntoType type = TypeCache.getTypeForName(node.getType().getInternalName());
        Operand res = getNewTemp(PredefinedType.Int8);
        emitter.emit("chkt %0 %1 %2", res, toCheck, type);
        environment.pushOperand(res);
    }

    @Override
    public void visit(FieldAccessExp node, Object arg) {
        Operand target = null;
        ClassBuilder cls;
        String memberName = node.getField().getIdentifier();
        node.getTarget().accept(this, arg);
        if (environment.peekOperand() != null) {
            target = environment.popOperand();
            if (target.isThis()) {
                cls = classBuilder;
                AttributeBuilder attribute = environment.lookupAttribute(memberName);
                if (attribute != null) {
                    OperandBuilder result = attribute.toOperand();
                    environment.pushOperand(new Operand(result, attribute.getType(), memberName));
                    return;
                }
            } else {
                IntoType type = target.getType();
                cls = ClassCache.getClassForName(type.getName());
                /////
                if (cls == null) {
                    logger.error("Not a class type: {name} at {node}", type.getName(), node.toString());
                    System.exit(1);
                }
                /////
                AttributeBuilder attribute = cls.getAttribute(memberName);
                if (attribute != null) {
                    int varSlot = target.getSlot();
                    int attrSlot = attribute.getSymbolSlot(false);
                    OperandBuilder qualifiedAttr =
                        OperandBuilder.objectFieldOperand(componentBuilder,
                                                          varSlot,
                                                          attrSlot);
                    environment.pushOperand(new Operand(qualifiedAttr,
                                                        attribute.getType(),
                                                        memberName));
                    return;
                }
            }
        } else {
            /////
            if (!(node.getTarget() instanceof IdExp)) {
                logger.error("Not a class: {class} at {node}", node.getTarget(), node.toString());
                System.exit(1);
            }
            /////
            String className = ((IdExp) node.getTarget()).getIdentifier();
            cls = ClassCache.getClassForName(className);
            /////
            if (cls == null) {
                logger.error("Not a class: {class} at {node}", className, node.toString());
                System.exit(1);
            }
            /////
            AttributeBuilder attribute = cls.getAttribute(memberName);
            if (attribute != null) {
                environment.pushOperand(new Operand(attribute.toOperand(),
                                                    attribute.getType(),
                                                    memberName));
                return;
            }
        }

        // must be a method
        MethodBuilder method = cls.getMethod(memberName);
        /////
        if (method == null) {
            logger.error("No such member: {name} at {node}", memberName, node.toString());
            System.exit(1);
        }
        /////
        ClassBuilder closure = componentBuilder.createClass("$Closure"+(__nextClosureID__++),
                                                            EnumSet.allOf(ClassModifier.class));
        IntoOperand attribute = null;
        boolean hasConstructor = false;
        if (target != null) {
            hasConstructor = true;
            MethodBuilder constructor = closure.createMethod("$init", EnumSet.of(MethodAccessFlags.Constructor));
            IntoType targetType = target.getType();
            String name = target.getName();
            if (target.isThis())
                name = "__this__";
            EnumSet<AttributeAccessFlags> accessFlags = EnumSet.noneOf(AttributeAccessFlags.class);
            attribute = closure.createAttribute(targetType, name, accessFlags);
            VariableBuilder argument = constructor.createArgument(targetType, name);
            BytecodeBuilder constructorBody = constructor.createBody();
            constructorBody.emitMove(attribute, argument);
        }
        MethodBuilder callee = closure.createMethod("call", EnumSet.noneOf(MethodAccessFlags.class));
        if (method.getArguments() != null) {
            method.getArguments().forEach(v -> callee.createArgument(v.getType(), v.getName()));
        }
        if (method.getReturnTypes() != null) {
            method.getReturnTypes().forEach(callee::addReturnType);
        }
        BytecodeBuilder calleeBody = callee.createBody();
        List<IntoOperand> calleeArgs = new ArrayList<>();
        List<IntoOperand> calleeResults = new ArrayList<>();
        callee.getArguments().forEach(calleeArgs::add);
        callee.getReturnTypes().forEach(v -> {
                VariableBuilder tmp = calleeBody.createVariable(v, "$tmp"+(__nextTempID__++));
                calleeResults.add(tmp);
            });
        if (attribute == null)
            calleeBody.emitCallStatic(method, cls, calleeArgs, calleeResults);
        else
            calleeBody.emitCallImmediate(attribute, method, cls, calleeArgs, calleeResults);
        if (calleeResults.size() > 0)
            calleeBody.emitReturn(calleeResults);
        else
            calleeBody.emitReturn();
        
        Operand dst = getNewTemp(closure.getClassType());
        emitter.emit("new %0 %1", dst, closure);
        if (hasConstructor) {
            MethodBuilder constructor = closure.getMethod("$init");
            emitter.emit("calli %0 %1 [%2] %3 %4", dst, constructor, target, new ArrayList<>(), closure);
        }
        environment.pushOperand(dst);
    }

    @Override
    public void visit(AnonFunctionExp node, Object arg) {
        String containingClassName = classBuilder.getName();
        ClassBuilder closure = componentBuilder.createClass(containingClassName + "_$Closure"+(__nextClosureID__++),
                                                            EnumSet.allOf(ClassModifier.class));
        MethodBuilder constructor = closure.createMethod("$init", EnumSet.of(MethodAccessFlags.Constructor));
        MethodBuilder callee = closure.createMethod("call", EnumSet.noneOf(MethodAccessFlags.class));
        List<Operand> captures = MapReduce.toList(node.getCaptures(), v -> {
                v.accept(this, arg);
                return environment.popOperand();
            });
        for (Operand capture : captures) {
            IntoType type = capture.getType();
            String name = capture.getName();
            if (capture.isThis())
                name = "__this__";
            EnumSet<AttributeAccessFlags> accessFlags = EnumSet.noneOf(AttributeAccessFlags.class);
            closure.createAttribute(type, name, accessFlags);
            constructor.createArgument(type, name);
        }
        BytecodeBuilder constructorBody = constructor.createBody();
        for (AttributeBuilder attribute : closure.getAttributes()) {
            VariableBuilder variable = constructor.getVariable(attribute.getName());
            constructorBody.emitMove(attribute, variable);
        }
        if (node.getParameters() != null) {
            node.getParameters().forEach(v -> {
                    IntoType type = TypeCache.getTypeForName(v.getType().getInternalName());
                    String name = v.getName();
                    callee.createArgument(type, name);
                });
        }
        if (node.getResultTypes() != null) {
            node.getResultTypes().forEach(v -> {
                    IntoType type = TypeCache.getTypeForName(v.getInternalName());
                    callee.addReturnType(type);
                });
        }
        node.getBody().accept(new MethodBodyGenerator(componentBuilder, closure, callee).withParent(this), arg);

        Operand dst = getNewTemp(closure.getClassType());
        emitter.emit("new %0 %1", dst, closure);
        emitter.emit("calli %0 %1 %2 %3 %4", dst, constructor, captures, new ArrayList<IntoOperand>(), closure);
        environment.pushOperand(dst);
    }


    // Helpers

    private boolean needsTrunc(IntoType lhs, IntoType rhs, boolean isOrdered) {
        boolean needsTrunc = false;
        if (rhs == null) {
            // special case for the null literal which currently has no type
            return false;
        }
        if (((rhs.getProtocolType() == TypeProtocol.Float32) ||
             (rhs.getProtocolType() == TypeProtocol.Float64)) &&
            (rhs.getProtocolType() != lhs.getProtocolType())) {
            needsTrunc = true;
        }
        if (isOrdered)
            return needsTrunc;
        return needsTrunc || needsTrunc(rhs, lhs, true);
    }

    private <L extends Node> void visitReverse(List<L> list, Object arg) {
        if (list != null) {
            ListIterator<L> it = list.listIterator(list.size());
            while (it.hasPrevious()) {
                it.previous().accept(this, arg);
            }
        }
    }

    private Operand getNewTemp(IntoType type) {
        String name = "$tmp"+__nextTempID__++;
        VariableBuilder var = bodyBuilder.createVariable(type, name);
        return new Operand(var.toOperand(), type, null);
    }

    private boolean insideClosure() {
        return parent != null;
    }

    @SuppressWarnings("unchecked")
    private <I> List<I> makeList(I... args) {
        List<I> lst = new ArrayList<>();
        if (args != null) {
            Collections.addAll(lst, args);
        }
        return lst;
    }
}
