package cola.compiler;

import cosma.bytecode.builder.*;

import java.util.ArrayList;
import java.util.List;

public class Operand implements IntoOperand {
    private final OperandBuilder obuilder;
    private final IntoType type;
    private final String name;
    private final List<Operand> indices;

    Operand(OperandBuilder obuilder, IntoType type, String name) {
        this.obuilder = obuilder;
        this.type = type;
        this.name = name;
        this.indices = new ArrayList<>();
    }

    public IntoType getType() {
        return type;
    }

    public OperandBuilder toOperand() {
        return obuilder;
    }

    public boolean isThis() {
        return obuilder.is(OperandType.This);
    }

    public boolean isVariable() {
        return obuilder.is(OperandType.Variable);
    }

    public boolean isAttribute() {
        return obuilder.is(OperandType.ObjectField);
    }

    public String getName() {
        ////////
        assert name != null : "Try to get name of temporary";
        ////////
        return name;
    }

    public int getSlot() {
        if (isVariable()) {
            OperandBuilder.Variable var = obuilder.asVariable();
            return var.getSlot();
        } else if (isAttribute()) {
            OperandBuilder.ObjectField field = obuilder.asObjectField();
            return field.getAttributeSlot();
        }
        ////////
        assert false : "Operand::getSlot() : Not a variable or attribute";
        ////////
        return -1;
    }

    public boolean isTemporary() {
        return name == null;
    }

    public Operand withIndex(Operand index) {
        indices.add(index);
        return this;
    }

    public List<Operand> getIndices() {
        return indices;
    }

    public boolean isArrayAccess() {
        return !indices.isEmpty();
    }

    public int getDimensions() {
        if (type instanceof ArrayTypeBuilder) {
            ArrayTypeBuilder builder = (ArrayTypeBuilder) type;
            return builder.getNumDimensions();
        } else {
            return 0;
        }
    }

    public IntoType getReadType() {
        if (isArrayAccess()) {
            return type.getMethod("read").getReturnTypes().iterator().next();
        } else {
            return type;
        }
    }

    public Operand clone() {
        return new Operand(obuilder, type, name);
    }

    public static Operand makeConstInt32(int value, Builder builder) {
        OperandBuilder op = OperandBuilder.constantOperand(builder, value);
        return new Operand(op, PredefinedType.Int32, null);
    }

    public static Operand makeConstFloat32(float value, Builder builder) {
        OperandBuilder op = OperandBuilder.constantOperand(builder, value);
        return new Operand(op, PredefinedType.Float32, null);
    }

    public static Operand makeConstFloat64(double value, Builder builder) {
        OperandBuilder op = OperandBuilder.constantOperand(builder, value);
        return new Operand(op, PredefinedType.Float64, null);
    }

    public static Operand makeConstBool(boolean value, Builder builder) {
        OperandBuilder op = OperandBuilder.constantOperand(builder, value ? 1 : 0);
        return new Operand(op, PredefinedType.Int8, null);
    }

    public static Operand makeConstChar(int value, Builder builder) {
        OperandBuilder op = OperandBuilder.constantOperand(builder, value);
        return new Operand(op, PredefinedType.Int8, null);
    }
}
