package cola.compiler;

import cola.ast.ClassNode;
import cola.ast.ComponentNode;
import cola.ast.MethodNode;
import cola.ast.exp.AnonFunctionExp;
import cola.ast.exp.NewArrayExp;
import cola.ast.exp.TypeVerdictExp;
import cola.ast.stm.AssignStm;
import cola.ast.stm.BlockStm;
import cola.ast.stm.Stm;
import cola.ast.stm.VarStm;
import cola.ast.type.*;
import cola.ast.visitor.GenericVisitorAdapter;
import cosma.bytecode.TypeModifier;
import cosma.bytecode.builder.*;
import cosma.tools.logger.LogManager;
import cosma.tools.logger.Logger;

import java.util.EnumSet;

public class TypeGenerator extends GenericVisitorAdapter<Object, IntoType> {
    private final Builder componentBuilder;
    private final Logger logger = LogManager.getCurrentClassLogger();
    
    TypeGenerator(Builder componentBuilder) {
        this.componentBuilder = componentBuilder;
    }

    public IntoType visit(ComponentNode node, Object arg) {
        if (node.getTypedefs() != null)
            node.getTypedefs().forEach(v -> v.accept(this, arg));
        if (node.getClasses() != null)
            node.getClasses().forEach(v -> v.accept(this, arg));
        return null;
    }

    public IntoType visit(ClassNode node, Object arg) {
        if (node.getAttributeNodes() != null)
            node.getAttributeNodes().forEach(v -> v.accept(this, arg));
        if (node.getMethods() != null)
            node.getMethods().forEach(v -> v.accept(this, arg));
        return null;
    }

    public IntoType visit(MethodNode node, Object arg) {
        if (node.getParameters() != null)
            node.getParameters().forEach(v -> v.accept(this, arg));
        if (node.getResultTypes() != null)
            node.getResultTypes().forEach(v -> v.accept(this, arg));
        if (node.getBody() != null)
            node.getBody().accept(this, arg);
        return null;
    }

    public IntoType visit(NewArrayExp node, Object arg) {
        node.getType().accept(this, arg);
        return null;
    }

    public IntoType visit(BlockStm node, Object arg) {
        for (Stm stm : node.getStms()) {
            stm.accept(this, arg);
        }
        return null;
    }

    public IntoType visit(VarStm node, Object arg) {
        if (node.getType() != null) {
            node.getType().accept(this, arg);
        }
        if (node.getInitializer() != null) {
            node.getInitializer().accept(this, arg);
        }
        return null;
    }

    public IntoType visit(AssignStm node, Object arg) {
        node.getLHS().accept(this, arg);
        node.getRHS().accept(this, arg);
        return null;
    }

    public IntoType visit(AnonFunctionExp node, Object arg) {
        if (node.getParameters() != null) {
            node.getParameters().forEach(v -> v.getType().accept(this, arg));
        }
        if (node.getResultTypes() != null) {
            node.getResultTypes().forEach(v -> v.accept(this, arg));
        }
        node.getBody().accept(this, arg);
        return null;
    }

    public IntoType visit(TypeVerdictExp node, Object arg) {
        node.getType().accept(this, arg);
        return null;
    }

    public IntoType visit(ArrayType node, Object arg) {
        String typeName = node.getInternalName();
        IntoType typeObject = TypeCache.getTypeForName(typeName);
        if (typeObject == null) {
            IntoType elementType = node.getElementType().accept(this, arg);
            typeObject = componentBuilder.getArrayType(elementType,
                                                       node.getDimensions(),
                                                       convertAssertions(node));
            TypeCache.addType(typeName, typeObject);
        }
        return typeObject;
    }

    public IntoType visit(StructuralType node, Object arg) {
        String definedName = null;
        if (arg instanceof String) {
            definedName = (String)arg;
            arg = null;
        }
        String typeName = node.getInternalName();
        IntoType typeObject = TypeCache.getTypeForName(typeName);
        if (typeObject == null) {
            TypeBuilder typeBuilder = componentBuilder.createType(definedName == null ? typeName : definedName);
            if (node.hasAncestor()) {
                IntoType ancestor = node.getAncestor().accept(this, arg);
                for (MethodBuilder mb : ancestor.getMethods()) {
                    MethodBuilder copy = mb.copyWithoutBody();
                    copy.setSemantic(PredefinedSemantic.Unspecified);
                    typeBuilder.addMethod(copy);
                }
            } else if (node.hasMethods()) {
                node.getMethods().forEach(m -> {
                        String methodName = m.getName();
                        MethodBuilder methodBuilder = typeBuilder.createMethod(methodName);
                        methodBuilder.setOptional(m.isOptional());
                        if (m.hasParameters()) {
                            m.getParameters().forEach(v -> {
                                    IntoType type = v.getType().accept(this, null);
                                    String name = v.getName();
                                    methodBuilder.createArgument(type, name, v.isSafe());
                                });
                        }
                        if (m.hasResults()) {
                            m.getResultTypes().forEach(v -> {
                                    IntoType type = v.accept(this, null);
                                    methodBuilder.addReturnType(type);
                                });
                        }
                        methodBuilder.setSemantic(PredefinedSemantic.Unspecified);
                    });
            } else {
                logger.debug("Ref type: {node}", node.getInternalName());
            }
            typeBuilder.setModifiers(convertAssertions(node));
            TypeCache.addType(typeName, typeBuilder);
            typeObject = typeBuilder;
        }
        return typeObject;
    }

    public IntoType visit(FunctionalType node, Object arg) {
        String typeName = node.getInternalName();
        IntoType typeObject = TypeCache.getTypeForName(typeName);
        if (typeObject == null) {
            TypeBuilder typeBuilder = componentBuilder.createType(typeName);
            typeBuilder.setModifiers(convertAssertions(node));
            MethodBuilder methodBuilder = typeBuilder.createMethod("call");
            if (node.hasParameters()) {
                node.getParameterTypes().forEach(v -> {
                        IntoType type = v.accept(this, arg);
                        methodBuilder.createArgument(type, "_");
                    });
            }
            if (node.hasResults()) {
                node.getResultTypes().forEach(v -> {
                        IntoType type = v.accept(this, arg);
                        methodBuilder.addReturnType(type);
                    });
            }
            TypeCache.addType(typeName, typeBuilder);
            typeObject = typeBuilder;
        }
        return typeObject;
    }

    public IntoType visit(TypeIdentifier node, Object arg) {
        String typeName = node.getInternalName();
        IntoType typeObject = TypeCache.getTypeForName(typeName);
        if (typeObject == null) {
            if (typeName.equals("String")) {
                IntoType elementType = PredefinedType.Int8;
                typeObject = componentBuilder.getArrayType(elementType,
                                                           1,
                                                           convertAssertions(node));
                TypeCache.addType("String", typeObject);
            } else {
                IntoType rawType = TypeCache.getTypeForName(node.getName());
                if (rawType != null) {
                    CopiedTypeBuilder copy = rawType.createCopy(componentBuilder, typeName);
                    copy.setModifiers(convertAssertions(node));
                    typeObject = copy;
                    TypeCache.addType(typeName, typeObject);
                } else {
                    typeObject = new ForwardDeclaredTypeBuilder(componentBuilder);
                    TypeCache.addType(typeName, typeObject);
                }
            }
        }
        return typeObject;
    }

    public IntoType visit(SimpleType node, Object arg) {
        String typeName = node.getInternalName();
        IntoType typeObject = TypeCache.getTypeForName(typeName);
        if (typeObject == null) {
            switch(typeName) {
            case "int":
            case "int32":
                typeObject = PredefinedType.Int32;
                break;
            case "bool":
            case "byte":
            case "int8":
                typeObject = PredefinedType.Int8;
                break;
            case "int16":
                typeObject = PredefinedType.Int16;
                break;
            case "int64":
                typeObject = PredefinedType.Int64;
                break;
            case "float":
                typeObject = PredefinedType.Float32;
                break;
            case "double":
                typeObject = PredefinedType.Float64;
                break;
            case "ref":
                typeObject = PredefinedType.Ref;
                break;
            default:
                throw new RuntimeException("'" + typeName + "' is not a valid type name");
            }
            TypeCache.addType(typeName, typeObject);
        }
        return typeObject;
    }

    private EnumSet<TypeModifier> convertAssertions(AbstractType type) {
        EnumSet<TypeModifier> res = EnumSet.noneOf(TypeModifier.class);
        if (type.isLocal()) res.add(TypeModifier.Local);
        if (type.isFinal()) res.add(TypeModifier.DefaultUnavail);
        if (!type.isTransient()) {
            res.add(TypeModifier.Storable);
            res.add(TypeModifier.AssertStorableState);
        }
        if (!type.isIntimate()) {
            res.add(TypeModifier.Transferable);
            res.add(TypeModifier.AssertTransferableState);
        }
        return res;
    }
}
