/*
 * COSMA -- A Virtual Machine for the Secure Cooperation of Untrusted Components.
 * Copyright (C) 2019  Roland Wismüller and Damian Ludwig, University of Siegen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package cola.util;

import java.util.*;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Collectors;

public abstract class MapReduce {
    /**
     * Maps a list of Ts to a list of Rs.
     *
     * @param l the list to map
     * @param f the mapper function
     * @return a list of elements returned by f, or empty list, if l is empty or null.
     */
    public static <R,T> List<R> toList(List<T> l, Function<? super T, ? extends R> f) {
        if (l == null || l.isEmpty()) {
            return new ArrayList<>();
        }
        return l.stream().map(f).collect(Collectors.toList());
    }

    /**
     * Maps a list of Ts to a set of Rs.
     *
     * @param l the list to map
     * @param f the mapper function
     * @return a set of elements returned by f, or empty set, if l is empty or null.
     */
    public static <R,T> Set<R> toSet(List<T> l, Function<? super T, ? extends R> f) {
        if (l == null || l.isEmpty()) {
            return new HashSet<>();
        }
        return l.stream().map(f).collect(Collectors.toSet());
    }

    /**
     * Maps a list of Ts to a comma separated string.
     *
     * @param l the list to map
     * @param f the mapper function
     * @return a comma separated string of the strings returned by f,
     * or empty string, if l is empty or null.
     */
    public static <T> String toString(List<T> l, Function<? super T, ? extends String> f) {
        if (l == null || l.isEmpty()) {
            return "";
        }
        return l.stream().map(f).collect(Collectors.joining(","));
    }

    /**
     * Queries the list for the first element matching the predicate.
     *
     * @param l the list to query
     * @param f the predicate function
     * @return the first element matching f or null
     */
    public static <T> T search(List<T> l, Predicate<? super T> f) {
        if (l == null || l.isEmpty()) {
            return null;
        }
        Optional<T> result = l.stream().filter(f).findFirst();
        return result.orElse(null);
    }
}
