/*
 * COSMA -- A Virtual Machine for the Secure Cooperation of Untrusted Components.
 * Copyright (C) 2019  Roland Wismüller and Damian Ludwig, University of Siegen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package cola.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Utility class to compute evaluation order for directed acyclic
 * graphs.
 */
public class TSort {
    private final int numNodes;
    private final int[] indegree;
    private final boolean[] marked;
    private final Map<Integer, List<Integer>> neighbours;

    public TSort(int num) {
        numNodes = num;
        indegree = new int[numNodes];
        marked = new boolean[numNodes];
        neighbours = new HashMap<>();
        
        for (int i = 0; i< numNodes; i++) {
            indegree[i] = 0;
            marked[i] = false;
            neighbours.put(i, new ArrayList<>());
        }
    }

    /**
     * Adds a relation to the graph.
     *
     * @param left needs to be processed before first
     * @param right needs 'left'
     */
    public void addRelation(Integer left, Integer right) {
        neighbours.get(left).add(right);
        indegree[right]++;
    }

    /**
     * Brings the nodes in any valid topological sorting
     *
     * @return the sorting as list, or null if the graph is cyclic
     */
    public List<Integer> sort() {
        List<Integer> list = new ArrayList<>();
        boolean isCyclic = false;

        while (list.size() < numNodes && !isCyclic) {
            isCyclic = true;
            for (int w = 0; w< numNodes; w++) {
                if (!marked[w] && indegree[w]==0) {
                    marked[w] = true;
                    isCyclic = false;
                    for (int k : neighbours.get(w))
                        indegree[k]--;
                        
                    list.add(w);
                }
            }
        }

        if (isCyclic)
            return null;

        return list;
    }
}
